#pragma once

#include <QtOpenGL>
#include <QVector>

#include <stdio.h>
#include <float.h>

#include "algebra3.h"
#include "tnbframe.h"

using namespace std;

class OFFFile {

public:

    OFFFile();
    ~OFFFile();

    bool Read(const QString & filename);
    bool Read_OBJ(const QString & filename);
    bool Write(const QString & filename) const;
    bool Write_OBJ(const QString & filename) const;

    void Clear();
    void Normalize();

    void DrawGL();
    void DrawCurvatureGL(const QList <float> & curvs);

    vec3 GetVert(const int i);
    QVector <vec3> & GetVerts();
    const QVector <vec3> & GetVerts() const;
    QVector <int> GetFace(const int i);
    QVector <QVector <int> > & GetFaces();
    const QVector <QVector <int> > & GetFaces() const;
    QVector <vec3> & GetFaceNorms();
    const QVector <vec3> & GetFaceNorms() const;

    vec3 GetFaceCentroid(const int face_index) const;

    //building geometry
    void AddCube(const vec3 & v, const float size);    
    void AddDisc(const vec3 & v, const vec3 & n, const float radius);
    void AddFace(const vec3 & v1, const vec3 & v2, const vec3 & v3);
    void AddFace(const vec3 & v1, const vec3 & v2, const vec3 & v3, const vec3 & v4);

    void ComputeFaceNormals();
    void FlipFaceNormals();

    float ComputeSurfaceArea(const bool special_case = false);

    int GetClosestVertexIndex(const vec3 & p) const;

    void ApplyModelviewMatrix(const float mview []);

private:

    void SetColorByCurvature(const vec2 curv, const float max_curv);

    QVector <vec3> m_vVertices; //vertices
    QVector <vec3> m_vVertNorms; //per-vertex normals
    QVector <vec3> m_vFaceNorms; //per-face normals

    QVector <QVector <int> > m_vFace_v; //indexes to verts for each face
    QVector <QVector <int> > m_vFace_n; //indexes to norms for each face

    GLuint disp_list;

};

