#ifndef COLESTUDYDATA_H
#define COLESTUDYDATA_H

#include <QtCore>
#include <QtOpenGL>
#include <GL/glu.h>

#include "algebra3.h"
#include "OFFFile.h"
#include "gaugemanipulator.h"
#include "basrelief.h"

struct GroundTruth {

    int index;
    int group;
    int screen_x;
    int screen_y;
    float slant;
    float tilt;
    vec3 normal;
    vec3 position;

};

struct CameraSetting {

    float modelview[16];
    float fov;
    int screen_x; //viewport
    int screen_y;

};

class ColeStudyData {

public:

    ColeStudyData();    
    void Load();

    void Draw(const int model_index);
    int GetViewportWidth(const int model_index);
    int GetViewportHeight(const int model_index);

    static void SlantTiltToNormal(const float slant, const float tilt, vec3 & normal);
    static void NormalToSlantTilt(const vec3 & normal, float & slant, float & tilt);

    static void SetGaugeColourByError(const float error_deg, GaugeManipulator & g);

private:

    void LoadGroundTruths();
    void LoadAskOrders();
    void LoadSlantTilt();
    void LoadTimes();
    void LoadCameraSettings();
    void LoadModels();    
    void LoadCurvaturesVertex();
    void ComputeCurvaturesGauge();

    void LoadViewDepCurvaturesVertex();
    void ComputeViewDepCurvaturesGauge();

    void SetupGroundTruthsGauges();
    void SetupUserAverageGauges();

    void ComputeBasRelief();
    void SaveBasRelief();
    void LoadBasRelief();

    void ComputeGaugeNearestVertex();

    void SaveCSVCurvatureAverageError();
    void SaveCSVViewAngleAverageError();
    void SaveCSVViewDepCurvatureAverageError();
    void SaveCSVMultiDimensionPlot();

    QString base_dir;
    QVector <QString> model_names;
    QVector <QList <GroundTruth > > ground_truths;    
    QVector <QVector <QVector <int> > > ask_orders; //[model][gaugeindex][participantindex]
    QVector <QVector <QVector <vec2> > > slant_tilts;
    QVector <QVector <QVector <bool> > > slant_tilts_set;
    QVector <QVector <QVector <vec3> > > input_normals_BR;
    QVector <QVector <QVector <int> > > times;

    QVector <bool> using_camera_a;
    QVector <CameraSetting> camera_settings;

    QVector <OFFFile> models;

    QVector <float> gauge_scales;

    QVector <QList <GaugeManipulator > > ground_truths_gauges;
    QVector <QList <GaugeManipulator > > user_avg_gauges;    
    QVector <QList <GaugeManipulator > > user_avg_gauges_BR;

    //averaged error for each gauge
    QVector <QList <float> > gauge_avg_error;
    QVector <QList <float> > gauge_avg_error_BR;

    //averaged error across the entire model
    QVector <float> avg_error_degrees;
    QVector <float> avg_error_degrees_BR;

    QVector <QList <int> > gauge_nearest_vertex;

    QVector <QList <vec2> > curvatures_vertex;
    QVector <QList <vec2> > curvatures_gauges;

    QVector <QList <float> > viewdep_curvatures_vertex;
    QVector <QList <float> > viewdep_curvatures_gauges;

};

#endif // COLESTUDYDATA_H
