#include "gaugemanipulator.h"

QList <vec3> GaugeManipulator::colormap_pink;

GaugeManipulator::GaugeManipulator() :
    cent(0, 0, 0),
    scale(0.5f),
    scalecontrolfac(0.8f),
    state(GM_NONE),
    mouseover(GM_NONE)
{

    rotcol[0] = 1;
    rotcol[1] = 1;
    rotcol[2] = 1;

    rotdrawcol[0] = 50;
    rotdrawcol[1] = 200;
    rotdrawcol[2] = 50;

    alwaysdrawcol[0] = 250;
    alwaysdrawcol[1] = 250;
    alwaysdrawcol[2] = 250;

    time.start();

    if (colormap_pink.empty()) {

        colormap_pink.push_back(vec3(255, 255, 255));
        colormap_pink.push_back(vec3(250, 250, 239));
        colormap_pink.push_back(vec3(244, 244, 220));
        colormap_pink.push_back(vec3(238, 238, 201));
        colormap_pink.push_back(vec3(232, 232, 178));
        colormap_pink.push_back(vec3(225, 217, 170));
        colormap_pink.push_back(vec3(219, 201, 161));
        colormap_pink.push_back(vec3(212, 184, 152));
        colormap_pink.push_back(vec3(205, 166, 142));
        colormap_pink.push_back(vec3(198, 145, 132));
        colormap_pink.push_back(vec3(190, 120, 120));
        colormap_pink.push_back(vec3(172, 107, 107));
        colormap_pink.push_back(vec3(152, 93, 93));
        colormap_pink.push_back(vec3(129, 76, 76));
        colormap_pink.push_back(vec3(100, 53, 53));
        colormap_pink.push_back(vec3(60, 0, 0));

    }

}

void GaugeManipulator::mousePressEvent(const QPoint & pos)
{

    state = mouseover;

    last_x = pos.x();
    last_y = pos.y();

}

void GaugeManipulator::mouseMoveEvent(const QPoint & pos, const QPoint & saved_pos, const OrbitCamera & cam)
{

    UpdatePickGL(pos.x(), pos.y());

    if (state == GM_ROT) {

        const int winWidth = GLutils::GetWindowWidth();
        int winHeight = GLutils::GetWindowHeight();

        const float ud_angle = -float(pos.y() - saved_pos.y()) / float(winHeight) * 200.0f;
        const float lr_angle = float(pos.x() - saved_pos.x()) / float(winWidth) * 200.0f;

        RotateNormalUD(cam, ud_angle);
        RotateNormalLR(cam, lr_angle);

    }

    /*
    //do rotation
    transplane.SetN(cam.ViewDir());
    transplane.SetD(transplane.N() * cent);

    vec3 p1;
    vec3 p2;

    GLutils::UnProjectPoint(vec2(x, y), 0.1f, p1);
    GLutils::UnProjectPoint(vec2(x, y), 0.2f, p2);

    vec3 newtransint;
    transplane.RayIntersect(p1, p2-p1, newtransint);

    mat3 rotMat = rotation3D(cam.ViewDir(), -90.0f);
    vec3 rotaxis = newtransint - cent;
    rotaxis = rotMat * rotaxis;
    rotaxis.normalize();

    float rotDelta = (newtransint - cent).length() - (transint - cent).length();
    rotDelta *= float(winWidth) * 0.2f;

    if (rotDelta < -5.0f) {
        rotDelta = -5.0f;
    }
    if (rotDelta > 5.0f) {
        rotDelta = 5.0f;
    }

    p.RotateFrame(rotaxis, rotDelta);

    transnorm = rotaxis;
    transint = newtransint;
    */

    last_x = pos.x();
    last_y = pos.y();

}

void GaugeManipulator::UpdatePickGL(int x, int y)
{

    //glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    DrawPickGL();

    unsigned char r, g, b;
    GLutils::ReadPixelColor_BackBuffer(x, y, r, g, b);

    if (r == rotcol[0] && g == rotcol[1] && b == rotcol[2]) {
        mouseover = GM_ROT;
    }    
    else {
        mouseover = GM_NONE;
    }

}

void GaugeManipulator::mouseReleaseEvent(const QPoint & pos)
{

    state = GM_NONE;

    last_x = pos.x();
    last_y = pos.y();

}

void GaugeManipulator::DrawGL(const float alpha)
{

    UpdateVerts();

    //float whiteval = ((sinf(time.elapsed() / 50.0f)) + 1.0f) * 0.25f;
    //vec3 whitecol = vec3(.5f + whiteval, .5f + whiteval, .75f + whiteval);
    vec3 whitecol = vec3(float(rotdrawcol[0])/255.0f + 0.4f,
                         float(rotdrawcol[1])/255.0f + 0.4f,
                         float(rotdrawcol[2])/255.0f + 0.4f);

    //draw some lines
    glLineWidth(1.0f);
    glColor4f(0, 0, 0, alpha);
    DrawAlwaysLines();

    //draw rot control
    glLineWidth(1.0f);
    if (state == GM_ROT || (state == GM_NONE && mouseover == GM_ROT)) {
        glColor4f(whitecol[0], whitecol[1], whitecol[2], alpha);
    }
    else {
        glColor4ub(rotdrawcol[0], rotdrawcol[1], rotdrawcol[2], alpha * 255);
    }
    glBegin(GL_TRIANGLE_FAN);
    DrawRotLines();
    glEnd();

    //visual debugging
    /*
    glColor3f(1, 0, 0);
    glBegin(GL_POINTS);
    glVertex3f(transint[0], transint[1], transint[2]);
    glEnd();

    glBegin(GL_LINES);
    glVertex3f(cent[0], cent[1], cent[2]);
    glVertex3f(cent[0] + transnorm[0] * scale, cent[1] + transnorm[1] * scale, cent[2] + transnorm[2] * scale);
    glEnd();
    transplane.DrawGL();
    */

    glLineWidth(1.0f);
    glColor3f(1,1,1);

}

void GaugeManipulator::DrawLineGL()
{

    UpdateVerts();

    glColor3ub(rotdrawcol[0], rotdrawcol[1], rotdrawcol[2]);

    glBegin(GL_LINES);
    glVertex3f(v1[0], v1[1], v1[2]);
    glVertex3f(v2[0], v2[1], v2[2]);
    glEnd();

}

void GaugeManipulator::DrawPickGL()
{

    UpdateVerts();  

    GLUquadric * quad = gluNewQuadric();

    glPushMatrix();

    glColor3ub(rotcol[0], rotcol[1], rotcol[2]);
    glTranslatef(v1[0], v1[1], v1[2]);
    glScalef(scale, scale, scale);
    gluSphere(quad, 2.0, 15, 15);
    glColor3f(1,1,1);

    glPopMatrix();

    gluDeleteQuadric(quad);

}

GaugeManipState GaugeManipulator::GetState()
{

    return state;

}

vec3 GaugeManipulator::GetCentre()
{
    return cent;
}

void GaugeManipulator::SetCentre(const vec3 & p)
{

    cent = p;    

}

void GaugeManipulator::SetNormal(const vec3 & n)
{

    normal = n;

}

vec3 GaugeManipulator::GetNormal()
{

    return normal;

}

void GaugeManipulator::RotateNormalUD(const OrbitCamera & cam, const float ud_angle)
{

    vec3 up = cam.Up();
    vec3 right = cam.ViewDir() ^ up;

    mat3 updownrot = identity2D();

    updownrot.rotation3D(right, ud_angle);

    normal = updownrot * normal;
    normal.normalize();

}

void GaugeManipulator::RotateNormalLR(const OrbitCamera & cam, const float lr_angle)
{

    vec3 up = cam.Up();

    mat3 leftrightrot = identity2D();

    leftrightrot.rotation3D(up, lr_angle);

    normal = leftrightrot * normal;
    normal.normalize();

}

void GaugeManipulator::SetScale(const float f)
{
    scale = f;
}

float GaugeManipulator::GetScale()
{
    return scale;
}

void GaugeManipulator::DrawRotLines()
{

    const vec3 B = GetOrthoVec(normal);
    const vec3 T = normal ^ B;

    for (float i=0.0f; i<ALG_M2_PI; i+=0.1f) {

        const vec3 v = cent + T * sinf(i) * scale + B * cosf(i) * scale;
        glVertex3f(v[0], v[1], v[2]);

    }

}

void GaugeManipulator::DrawAlwaysLines()
{

    glBegin(GL_LINE_LOOP);
    DrawRotLines();
    glEnd();

    glBegin(GL_LINES);
    glVertex3f(v1[0], v1[1], v1[2]);
    glVertex3f(v2[0], v2[1], v2[2]);
    glEnd();

}

void GaugeManipulator::UpdateVerts()
{   

    v1 = cent;
    v2 = cent + normal * scale;

}

GLubyte GaugeManipulator::GetColourR()
{
    return rotdrawcol[0];
}

GLubyte GaugeManipulator::GetColourG()
{
    return rotdrawcol[1];
}

GLubyte GaugeManipulator::GetColourB()
{
    return rotdrawcol[2];
}

void GaugeManipulator::SetColour(const GLubyte r, const GLubyte g, const GLubyte b)
{

    rotdrawcol[0] = r;
    rotdrawcol[1] = g;
    rotdrawcol[2] = b;

}

void GaugeManipulator::SetColourByError(const float error_deg)
{

    /*
    float val = error_deg;
    val -= 15.0f;
    val = qMax(val, 0.0f);
    val /= 45.0f;
    val *= 255.0f;

    GLubyte col = GLubyte(val);

    rotdrawcol[0] = 255;
    rotdrawcol[1] = 255 - col;
    rotdrawcol[2] = 255 - col;
    */

    float error_deg2 = qMin(error_deg - 15.0f, 60.0f);
    error_deg2 = qMax(error_deg2, 0.0f);

    int val = int(error_deg2 / 60.0f * float(colormap_pink.size()));
    val = qMin(val, colormap_pink.size()-1);
    rotdrawcol[0] = GLubyte(colormap_pink[val][0]);
    rotdrawcol[1] = GLubyte(colormap_pink[val][1]);
    rotdrawcol[2] = GLubyte(colormap_pink[val][2]);

}
