#ifndef GAUGEMANIPULATOR_H
#define GAUGEMANIPULATOR_H

#include <QtOpenGL>

#include "algebra3.h"
#include "glutils.h"
#include "orbitcamera.h"
#include "plane.h"

enum GaugeManipState
{
    GM_NONE,
    GM_ROT
};

class GaugeManipulator
{

public:

    GaugeManipulator();

    void mousePressEvent(const QPoint & pos);
    void mouseMoveEvent(const QPoint & pos, const QPoint & saved_pos, const OrbitCamera & cam);
    void mouseReleaseEvent(const QPoint & pos);

    void DrawGL(const float alpha);
    void DrawLineGL();

    GaugeManipState GetState();

    vec3 GetCentre();
    void SetCentre(const vec3 & p);

    void SetNormal(const vec3 & n);
    vec3 GetNormal();

    void RotateNormalUD(const OrbitCamera & cam, const float ud_angle);
    void RotateNormalLR(const OrbitCamera & cam, const float lr_angle);

    void SetScale(const float f);
    float GetScale();

    GLubyte GetColourR();
    GLubyte GetColourG();
    GLubyte GetColourB();
    void SetColour(const GLubyte r, const GLubyte g, const GLubyte b);
    void SetColourByError(const float error_deg);

private:

    void DrawPickGL();
    void UpdatePickGL(int x, int y);

    void DrawRotLines();
    void DrawAlwaysLines();

    void UpdateVerts();

    //TNBFrame p; //plane implemented as tnb frame (normal is N)
    vec3 normal;
    vec3 cent;

    float scale;
    float scalecontrolfac;

    GLubyte rotcol[3];
    GLubyte rotdrawcol[3];

    GLubyte alwaysdrawcol[3];

    int last_x;
    int last_y;

    GaugeManipState state;
    GaugeManipState mouseover;

    QTime time;

    vec3 v1;
    vec3 v2;

    //for visual debugging
    Plane transplane;
    vec3 transint;
    vec3 transnorm;

    static QList <vec3> colormap_pink;

};

#endif // GAUGEMANIPULATOR_H
