#include "glutils.h"

int GLutils::point_disp_list = 0;

GLutils::GLutils()
{
}

void GLutils::DrawPointGL(const float x, const float y, const float rad)
{

    glPushMatrix();

    glTranslatef(x, y, 0);
    glScalef(rad, rad, 1);

    if (point_disp_list > 0) {
        glCallList(point_disp_list);
    }
    else {

        point_disp_list = glGenLists(1);
        glNewList(point_disp_list, GL_COMPILE_AND_EXECUTE);

        glBegin(GL_TRIANGLE_FAN);

        glVertex2f(0, 0);
        for (int i=0; i<=50; ++i) {

            float radval = 3.14159f * 2.0f * float(i) / float(50);

            glVertex2f(sinf(radval), cosf(radval));

        }

        glEnd();

        glEndList();
    }

    glPopMatrix();

}

void GLutils::SetColorByHue(const float val, const float range)
{

    float hue =  val / range * 255.0f;
    QColor col = QColor::fromHsv(hue, 255, 255);
    glColor3f(col.redF(), col.greenF(), col.blueF());

}

vec3 GLutils::ColorByHue(const float val, const float range)
{
    float hue =  val / range * 255.0f;
    QColor col = QColor::fromHsv(hue, 255, 255);
    return vec3(col.redF(), col.greenF(), col.blueF());
}

void GLutils::ColorByNearestMajorAxis(const vec3 & v)
{

    vec3 axes[6];
    axes[0] = vec3(1, 0, 0);
    axes[1] = vec3(0, 1, 0);
    axes[2] = vec3(0, 0, 1);
    axes[3] = vec3(-1, 0, 0);
    axes[4] = vec3(0, -1, 0);
    axes[5] = vec3(0, 0, -1);

    ColorByIndex(6);

    for (int i=0; i<6; ++i) {
        if (v == axes[i]) {
            ColorByIndex(i);
        }
    }

}

void GLutils::ColorByIndex(const int ind)
{

    int numcol = 7;

    if (ind % numcol == 0) {
        glColor3ub(61, 83, 140);
    }
    else if (ind % numcol == 1) {
        glColor3ub(249, 117, 114);
    }
    else if (ind % numcol == 2) {
        glColor3ub(159, 21, 141);
    }
    else if (ind % numcol == 3) {
        glColor3ub(255, 135, 8);
    }
    else if (ind % numcol == 4) {
        glColor3ub(19, 198, 127);
    }
    else if (ind % numcol == 5) {
        glColor3ub(222, 221, 120);
    }
    else if (ind % numcol == 6) {
        glColor3ub(242, 4, 8);
    }

}

void GLutils::DrawTNBFrame(const vec3 & o, const TNBFrame & tnb, const float scale)
{

    vec3 t = o + tnb.T() * scale;
    vec3 n = o + tnb.N() * scale;
    vec3 b = o + tnb.B() * scale;

    glBegin(GL_LINES);

    glColor3f(1,0,0);
    glVertex3d(o[0], o[1], o[2]);
    glVertex3d(t[0], t[1], t[2]);

    glColor3f(0,1,0);
    glVertex3d(o[0], o[1], o[2]);
    glVertex3d(n[0], n[1], n[2]);

    glColor3f(0,0,1);
    glVertex3d(o[0], o[1], o[2]);
    glVertex3d(b[0], b[1], b[2]);

    glEnd();

}

vec2 GLutils::ProjectPoint(const vec3 & p)
{

    GLdouble model_view[16];
    GLdouble projection[16];
    GLint viewport[4];
    GLdouble winx, winy, winz;

    glGetDoublev(GL_MODELVIEW_MATRIX, model_view);
    glGetDoublev(GL_PROJECTION_MATRIX, projection);
    glGetIntegerv(GL_VIEWPORT, viewport);

    gluProject(p[0], p[1], p[2], model_view, projection, viewport, &winx, &winy, &winz);

    return vec2(winx, winy);

}

vec3 GLutils::ProjectPoint3D(const vec3 & p)
{

    GLdouble model_view[16];
    GLdouble projection[16];
    GLint viewport[4];
    GLdouble winx, winy, winz;

    glGetDoublev(GL_MODELVIEW_MATRIX, model_view);
    glGetDoublev(GL_PROJECTION_MATRIX, projection);
    glGetIntegerv(GL_VIEWPORT, viewport);

    gluProject(p[0], p[1], p[2], model_view, projection, viewport, &winx, &winy, &winz);

    return vec3(winx, winy, winz);

}

QVector <vec2> GLutils::ProjectPoints(const QVector <vec3> & pts)
{

    GLdouble model_view[16];
    GLdouble projection[16];
    GLint viewport[4];
    GLdouble winx, winy, winz;

    glGetDoublev(GL_MODELVIEW_MATRIX, model_view);
    glGetDoublev(GL_PROJECTION_MATRIX, projection);
    glGetIntegerv(GL_VIEWPORT, viewport);

    QVector <vec2> proj_pts(pts.size());

    for (int i=0; i<pts.size(); ++i) {
        gluProject(pts[i][0], pts[i][1], pts[i][2], model_view, projection, viewport, &winx, &winy, &winz);
        proj_pts[i] = vec2(winx, winy);
    }

    return proj_pts;

}

vec2 GLutils::ProjectVector(const vec3 & p1, const vec3 & p2)
{

    vec2 tangent = GLutils::ProjectPoint(p2) - GLutils::ProjectPoint(p1);
    return tangent;

}

vec2 GLutils::ProjectUnitVector(const vec3 & p1, const vec3 & p2)
{

    vec2 tangent = GLutils::ProjectPoint(p2) - GLutils::ProjectPoint(p1);
    tangent.normalize();
    return tangent;

}

void GLutils::ProjectCurve(const QList <vec4> & c, QList <vec4> & proj_c)
{

    GLdouble model_view[16];
    GLdouble projection[16];
    GLint viewport[4];
    GLdouble winx, winy, winz;

    glGetDoublev(GL_MODELVIEW_MATRIX, model_view);
    glGetDoublev(GL_PROJECTION_MATRIX, projection);
    glGetIntegerv(GL_VIEWPORT, viewport);

    proj_c.clear();
    for (int i=0; i<c.size(); ++i) {

        gluProject(c[i][0], c[i][1], c[i][2], model_view, projection, viewport, &winx, &winy, &winz);
        proj_c.push_back(vec4(winx, winy, 0.0, c[i][3]));

    }

}

void GLutils::UnProjectPoint(const vec2 & p, const float depth_value, vec3 & unproj_pt)
{

    GLdouble model_view[16];
    GLdouble projection[16];
    GLint viewport[4];
    GLdouble objx, objy, objz;

    glGetDoublev(GL_MODELVIEW_MATRIX, model_view);
    glGetDoublev(GL_PROJECTION_MATRIX, projection);
    glGetIntegerv(GL_VIEWPORT, viewport);

    gluUnProject(p[0], p[1], depth_value, model_view, projection, viewport, &objx, &objy, &objz);

    unproj_pt = vec3(objx, objy, objz);

}

void GLutils::UnProjectCurve(const QList <vec2> & c, const QVector <float> depth_values, QList <vec3> & unproj_curve)
{

    unproj_curve.clear();

    GLdouble model_view[16];
    GLdouble projection[16];
    GLint viewport[4];
    GLdouble objx, objy, objz;

    glGetDoublev(GL_MODELVIEW_MATRIX, model_view);
    glGetDoublev(GL_PROJECTION_MATRIX, projection);
    glGetIntegerv(GL_VIEWPORT, viewport);

    for (int i=0; i<c.size(); ++i) {

        gluUnProject(c[i][0], c[i][1], depth_values[i], model_view, projection, viewport, &objx, &objy, &objz);

        unproj_curve.push_back(vec3(objx, objy, objz));

    }

}

void GLutils::EnableBlending()
{
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
}

void GLutils::DisableBlending()
{
    glDisable(GL_BLEND);
}

void GLutils::ReadPixelColor(const int x, const int y, unsigned char & r, unsigned char & g, unsigned char & b)
{

    glReadPixels(x, y, 1, 1, GL_RED, GL_UNSIGNED_BYTE, &r);
    glReadPixels(x, y, 1, 1, GL_GREEN, GL_UNSIGNED_BYTE, &g);
    glReadPixels(x, y, 1, 1, GL_BLUE, GL_UNSIGNED_BYTE, &b);

}

void GLutils::ReadPixelColor_FrontBuffer(const int x, const int y, unsigned char & r, unsigned char & g, unsigned char & b)
{

    glReadBuffer(GL_FRONT);
    ReadPixelColor(x, y, r, g, b);

}

void GLutils::ReadPixelColor_BackBuffer(const int x, const int y, unsigned char & r, unsigned char & g, unsigned char & b)
{

    glReadBuffer(GL_BACK);
    ReadPixelColor(x, y, r, g, b);

}

int GLutils::GetWindowWidth()
{

    GLint viewport[4];
    glGetIntegerv(GL_VIEWPORT, viewport);
    return viewport[2];

}

int GLutils::GetWindowHeight()
{

    GLint viewport[4];
    glGetIntegerv(GL_VIEWPORT, viewport);
    return viewport[3];

}

void GLutils::GetViewport(int v[4])
{

    glGetIntegerv(GL_VIEWPORT, v);

}

void GLutils::SetViewport(const int v[4])
{

    glViewport(v[0], v[1], v[2], v[3]);

}

void GLutils::SetViewport(const int x, const int y, const int w, const int h)
{

    glViewport(x, y, w, h);

}
