#ifndef GLUTILS_H
#define GLUTILS_H

#include <QtOpenGL>
#include <QtGui>
#include <GL/glu.h>

#include "algebra3.h"
#include "tnbframe.h"

class GLutils
{

public:

    GLutils();

    static void DrawPointGL(const float x, const float y, const float rad);
    static void SetColorByHue(const float val, const float range);
    static vec3 ColorByHue(const float val, const float range);
    static void ColorByIndex(const int ind);
    static void ColorByNearestMajorAxis(const vec3 & v);
    static void DrawTNBFrame(const vec3 & o, const TNBFrame & tnb, const float scale);
    static vec2 ProjectPoint(const vec3 & p);
    static vec3 ProjectPoint3D(const vec3 & p);
    static QVector <vec2> ProjectPoints(const QVector <vec3> & pts);
    static vec2 ProjectVector(const vec3 & p1, const vec3 & p2);
    static vec2 ProjectUnitVector(const vec3 & p1, const vec3 & p2);
    static void ProjectCurve(const QList <vec4> & c, QList <vec4> & proj_c);
    static void UnProjectPoint(const vec2 & p, const float depth_value, vec3 & unproj_pt);
    static void UnProjectCurve(const QList <vec2> & c, const QVector <float> depth_values, QList <vec3> & unproj_curve);
    static void EnableBlending();
    static void DisableBlending();

    static void ReadPixelColor_FrontBuffer(const int x, const int y, unsigned char & r, unsigned char & g, unsigned char & b);
    static void ReadPixelColor_BackBuffer(const int x, const int y, unsigned char & r, unsigned char & g, unsigned char & b);

    static int GetWindowWidth();
    static int GetWindowHeight();
    static void GetViewport(int v[4]);
    static void SetViewport(const int v[4]);
    static void SetViewport(const int x, const int y, const int w, const int h);

private:

    static void ReadPixelColor(const int x, const int y, unsigned char & r, unsigned char & g, unsigned char & b);

    static int point_disp_list;

};

#endif // GLUTILS_H
