#include "glwidget.h"

GLWidget::GLWidget() :
    shift_key_held(false),
    control_key_held(false),
    mouse_left(false),
    mouse_middle(false),
    mouse_right(false),
    study_numgaugespermodel(5),
    study_trialindex(-1),
    gauge_scale(0.018f), //0.04f
    num_halton_sequence_gauges(100),
    num_total_sequence_gauges(105),
    study_finished(false),    
    draw_all_gauges(false),
    experimental_mode(true), //NOTE: for Turk version set to false!
    load_and_show_study_data(false)
{   

    //stuff messing around for DGP logo for t-shirt :)
    /*
    OFFFile dgplogofile;
    dgplogofile.Read("dgp_logo_cntower.off");

    MedialAxis dgplogo;
    dgplogo.GetSurfaceList(dgplogofile.GetVerts(), dgplogofile.GetFaces(), 0.05f);
    dgplogo.WriteSurface("dgp_logo_cntower_cubie.off");
    */

    //coleStudyIndex = 0;      

    setMouseTracking(true);

    animate_timer.setInterval(15);
    animate_timer.start();
    connect(&animate_timer, SIGNAL(timeout()), this, SLOT(update()));

    const float default_phi = 60.0f;
    const float default_theta = 45.0f;
    const float default_cam_width = 1.35f;

    //meshnames.push_back("human");
    //meshnames.push_back("plane");
    meshnames.push_back("cup");
    meshnames.push_back("ant");
    meshnames.push_back("table");
    meshnames.push_back("camel_smooth");
    meshnames.push_back("cyl");
    meshnames.push_back("cyl_bumpy");

    meshes.resize(meshnames.size());
    meshstaticviews.resize(meshnames.size());
    meshcurv1s.resize(meshnames.size());
    meshcurv2s.resize(meshnames.size());

    for (int i=0; i<meshnames.size(); ++i) {

        //set the view
        if (QString::compare(meshnames[i], "human") == 0) {
            meshstaticviews[i].SetPhi(default_phi);
            meshstaticviews[i].SetTheta(default_theta + 160.0f);
            meshstaticviews[i].SetLookAt(vec3(0, -0.025, 0));
            meshstaticviews[i].SetCamWidth(default_cam_width);
            meshstaticviews[i].Update();
        }
        else if (QString::compare(meshnames[i], "cup") == 0) {
            meshstaticviews[i].SetPhi(default_phi - 10.0f);
            meshstaticviews[i].SetTheta(default_theta + 70.0f);
            meshstaticviews[i].SetLookAt(vec3(0, 0.1, 0));
            meshstaticviews[i].SetCamWidth(default_cam_width);
            meshstaticviews[i].Update();
        }
        else if (QString::compare(meshnames[i], "plane") == 0) {
            meshstaticviews[i].SetPhi(default_phi);
            meshstaticviews[i].SetTheta(default_theta + 85.0f);
            meshstaticviews[i].SetCamWidth(default_cam_width - 0.15f);
            meshstaticviews[i].Update();
        }
        else if (QString::compare(meshnames[i], "ant") == 0) {
            meshstaticviews[i].SetPhi(default_phi - 20.0f);
            meshstaticviews[i].SetTheta(default_theta  + 170.0f);
            meshstaticviews[i].SetCamWidth(default_cam_width - 0.15f);
            meshstaticviews[i].Update();
        }
        else if (QString::compare(meshnames[i], "table") == 0) {
            meshstaticviews[i].SetPhi(default_phi);
            meshstaticviews[i].SetTheta(default_theta + 70.0f);
            meshstaticviews[i].SetCamWidth(default_cam_width - 0.25f);
            meshstaticviews[i].Update();
        }
        else if (QString::compare(meshnames[i], "camel_smooth") == 0) {
            meshstaticviews[i].SetPhi(default_phi);
            meshstaticviews[i].SetTheta(default_theta - 20.0f);
            meshstaticviews[i].SetLookAt(vec3(0, -0.025, 0));
            meshstaticviews[i].SetCamWidth(default_cam_width);
            meshstaticviews[i].Update();
        }
        else if (QString::compare(meshnames[i], "cylinder") == 0 ||
                 QString::compare(meshnames[i], "cylinder_bumpy") == 0 ||
                 QString::compare(meshnames[i], "cyl") == 0 ||
                 QString::compare(meshnames[i], "cyl_bumpy") == 0) {
            meshstaticviews[i].SetPhi(default_phi - 5.0f);
            meshstaticviews[i].SetTheta(default_theta - 5.0f);
            meshstaticviews[i].SetLookAt(vec3(0, 0, 0));
            meshstaticviews[i].SetCamWidth(default_cam_width - 0.15f);
            meshstaticviews[i].Update();
        }        
        else {
            meshstaticviews[i].SetPhi(default_phi);
            meshstaticviews[i].SetTheta(default_theta);
            meshstaticviews[i].SetLookAt(vec3(0, 0, 0));
            meshstaticviews[i].SetCamWidth(default_cam_width);
            meshstaticviews[i].Update();
        }

    }    

    gaugemanip.SetScale(gauge_scale);
    cam.SetCamWidth(1.0f);         

    //coleStudyData.Load();

}

void GLWidget::SetupStudy(const QString & turker_id)
{

    //TODO: setup study based on turker_id...
    //BOTH the conditions, and also the random set of gauges shown (which are stored locally in a file)

    turk_worker_id = turker_id;

    //hashing for first run of study
    //hashed_id = qHash(turker_id) % 10; //10 conditions - 2 tasks * 5 abstractions

    //hashing for second run of study
    /*
    switch (qHash(turker_id) % 5) {
    case 0:
    case 1:
        hashed_id = 2;
        break;
    case 2:
    case 3:
        hashed_id = 4;
        break;
    case 4:
        hashed_id = 7;
        break;
    }
    qDebug() << turk_worker_id << "hashed condition" << hashed_id;
    */

    hashed_id = 6;

    study_task = (hashed_id < 5) ? ROTATE : ROTATE_VIEW;
    abstraction = hashed_id % 5; //0 original surface, 1 - MPS (planecuts), 2 - regxyz slices, 3 - radial slices, 4 - crdbrd
    //abstraction = 5;

    //load all the halton sequence stuff (some info is specific to condition, such as abstraction distance at a gauge)
    LoadHaltonSequences();

    if (load_and_show_study_data) {

        LoadTurkStudyDataFromCSV("./turk_study/data.csv", worker_data);
        worker_data.ComputeDataPoints(ground_truths, 6, 105, 10);
        worker_data.OutputStatistics(meshnames);

    }

    //load all the models specific to these conditions
    for (int i=0; i<meshnames.size(); ++i) {

        //load the model
        QString filename = QString("./input_models/") + meshnames[i] + QString("/");
        switch (abstraction) {
        case 0:
            filename += QString("inputmesh.obj"); break;
        case 1:
            filename += QString("slab.off"); break;
        case 2:
            filename += QString("regxyz.off"); break;
        case 3:
            filename += QString("radial.off"); break;
        case 4:
            filename += QString("crdbrd.off"); break;
        case 5:
            filename += QString("mpsmod.off"); break;
        //case 6:
        //    filename += QString("vsa80_subdiv.off"); break;
        }

        if (QString::compare(filename.right(3), "obj", Qt::CaseInsensitive) == 0) {
            meshes[i].Read_OBJ(filename);
        }
        else {
            meshes[i].Read(filename);
        }

    }

    LoadSplashImage();

    //setup random set of trials
    SetupAllTrials(); 

}

QSize GLWidget::minimumSizeHint() const
{
    return QSize(400, 300);
}

QSize GLWidget::sizeHint() const
{
    return QSize(800, 600);
}

void GLWidget::initializeGL()
{

    glEnable(GL_NORMALIZE);
    glEnable(GL_DEPTH_TEST);

    glEnable(GL_POINT_SMOOTH);
    glEnable(GL_LINE_SMOOTH);

    //glEnable(GL_BLEND);

    glDisable(GL_CULL_FACE);

    glShadeModel(GL_SMOOTH);

    glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);

    glClearColor(1, 1, 1, 1);      

}

void GLWidget::paintGL()
{

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);  //clears the colour and depth buffers

    if (study_trialindex == -1) {

        DrawSplashImage();

    }
    else {

        //coleStudyData.Draw(coleStudyIndex);
        //return;

        //draw in 3-space
        UpdateCamera();

        StudyData & d = studydatas[study_trialindex];

        meshes[d.model_index].DrawGL();

        //
        //vec3 v1 = d.pos;
        //vec3 v2 = d.pos + d.true_normal * gaugemanip.GetScale();

        //glColor3f(1, .5f, .5f);
        //glBegin(GL_LINES);
        //glVertex3f(v1[0], v1[1], v1[2]);
        //glVertex3f(v2[0], v2[1], v2[2]);
        //glEnd();

        //test medial axis
        //medialaxis.DrawGL();

        //debug draw intersection rays
        /*
        glColor3f(1, 0, 0);
        glBegin(GL_LINES);
        for (int i=0; i<ray1.size(); ++i) {

            glVertex3f(ray1[i][0], ray1[i][1], ray1[i][2]);
            glVertex3f(ray2[i][0], ray2[i][1], ray2[i][2]);

        }
        glEnd();
        */

        //return; //REMVOE THIS TO SEE GAUGES!!!
        glClear(GL_DEPTH_BUFFER_BIT);

        if (load_and_show_study_data) {

            for (int i=0; i<turk_studydata_gauges.size(); ++i) {

                turk_studydata_gauges[i].DrawGL(1.0f);

            }

            //draw all the samples
            //worker_data.DrawAllSamples(hashed_id, d.model_index, gauge_scale);

        }
        else {           

            if (experimental_mode && draw_all_gauges) {

                //for debugging the halton sequence/ground truth gauges:
                for (int i=0; i<ground_truths_gauges[d.model_index].size(); ++i) {
                    ground_truths_gauges[d.model_index][i].DrawGL(1.0f);
                }

            }
            else {

                const vec3 static_dir = meshstaticviews[studydatas[study_trialindex].model_index].ViewDir();
                const float cam_angle = acosf(cam.ViewDir() * static_dir) * RAD_TO_DEG;

                //if (cam_angle < 0.1f) {
                    gaugemanip.DrawGL(1.0f);
                //}

            }

            /*
        //this does gauge fading out when view deviates
        float cam_angle = acosf(cam.ViewDir() * meshstaticviews[studydatas[study_trialindex].model_index].ViewDir()) * RAD_TO_DEG;
        float gauge_alpha = 1.0f - cam_angle / 0.01f;
        gauge_alpha = qMin(gauge_alpha, 1.0f);
        gauge_alpha = qMax(gauge_alpha, 0.0f);

        glEnable(GL_BLEND);
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

        gaugemanip.DrawGL(gauge_alpha);

        glDisable(GL_BLEND);
        */

        }

        if (glGetError()) {
            qDebug() << "GLproblem";
        }

    }
}

void GLWidget::resizeGL(int width, int height)
{

    glViewport(0, 0, width, height);

}

void GLWidget::mousePressEvent(QMouseEvent *event)
{

    mouse_left = (event->buttons() & Qt::LeftButton);
    mouse_middle = (event->buttons() & Qt::MiddleButton);
    mouse_right = (event->buttons() & Qt::RightButton);

    if (study_finished) {
        return;
    }

    if (study_trialindex == -1) {

        SetupNewTrial();
        return;

    }

    makeCurrent();

    QPoint pos = mapFromGlobal(QCursor::pos());
    pos.setY(height() - pos.y());   

    if (mouse_left || mouse_right) {

        gaugemanip.mousePressEvent(pos);

        if (gaugemanip.GetState() != GM_NONE) {

            ++studydatas[study_trialindex].num_gauge_manipulations;

            mouse_saved_pos = pos;
            setCursor( QCursor( Qt::BlankCursor ) );

        }
        else {

            if (!study_finished && study_task == ROTATE_VIEW) {

                ++studydatas[study_trialindex].num_cam_manipulations;

            }

        }

    }

    mouse_pos = pos;

    updateGL();

}

void GLWidget::mouseMoveEvent(QMouseEvent *)
{

    if (study_finished) {
        return;
    }

    //vec2 pos(double(event->x()), double(height() - event->y()));
    QPoint pos = mapFromGlobal(QCursor::pos());
    pos.setY(height() - pos.y());

    if (gaugemanip.GetState() != GM_NONE && pos == mouse_saved_pos) {
        return;
    }

    //update mouseover for picking
    UpdateCamera();
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    gaugemanip.mouseMoveEvent(pos, mouse_saved_pos, cam);

    if (gaugemanip.GetState() == GM_NONE) {

        //camera manipulations only enabled when not doing static view study
        if (study_task == ROTATE_VIEW) {

            if (mouse_left || mouse_right) {

                if (shift_key_held) {

                    if (pos.y() < mouse_pos.y()) {
                        ZoomIn();
                    }
                    else if (pos.y() > mouse_pos.y()) {
                        ZoomOut();
                    }

                }
                else if (control_key_held) {
                    cam.Pan(pos.x() - mouse_pos.x(), pos.y() - mouse_pos.y());
                }
                else {

                    cam.Orbit(pos.x() - mouse_pos.x(), pos.y() - mouse_pos.y());
                    cam.SetInterpCamera(cam, 1.0f);

                }

            }

        }

    }
    else {

        emit EnableNextTrialButton();

        if (QCursor::pos() !=  mouse_saved_pos) {
            QPoint snap_back_pt = mouse_saved_pos;
            snap_back_pt.setY(height() - snap_back_pt.y());
            QCursor::setPos(mapToGlobal(snap_back_pt));
        }

    }

    mouse_pos = pos;

    updateGL();

}

void GLWidget::ZoomIn()
{

    //a zoomed in/out cam to interpolate to
    OrbitCamera newcam(cam);
    newcam.ZoomIn();
    cam.SetInterpCamera(newcam, 200.0f);

}

void GLWidget::ZoomOut()
{

    //a zoomed in/out cam to interpolate to
    OrbitCamera newcam(cam);
    newcam.ZoomOut();
    cam.SetInterpCamera(newcam, 200.0f);

}

void GLWidget::wheelEvent(QWheelEvent * e)
{  

    if (e->delta() > 0) {
        ZoomIn();
    }
    else if (e->delta() < 0) {
        ZoomOut();
    }   

}

void GLWidget::mouseReleaseEvent(QMouseEvent *event)
{

    mouse_left = (event->button() == Qt::LeftButton);
    mouse_middle = (event->button() == Qt::MiddleButton);
    mouse_right = (event->button() == Qt::RightButton);

    if (gaugemanip.GetState() != GM_NONE) {
        QPoint snap_back_pt = mouse_saved_pos;
        snap_back_pt.setY(height() - snap_back_pt.y());
        QCursor::setPos(mapToGlobal(snap_back_pt));
    }
    setCursor( QCursor( Qt::ArrowCursor ) );

    if (study_finished) {
        return;
    }

    makeCurrent();

    //vec2 pos(double(event->x()), double(height() - event->y()));
    QPoint pos = mapFromGlobal(QCursor::pos());
    pos.setY(height() - pos.y());   

    gaugemanip.mouseReleaseEvent(pos);

    if (gaugemanip.GetState() == GM_NONE) {

        OrbitCamera newcam = meshstaticviews[studydatas[study_trialindex].model_index];
        cam.SetInterpCamera(newcam, 750.0f);

    }

    if (draw_all_gauges && mouse_right) {

        //qDebug() << "mouse right";

        //find closest gauge to click
        float best_dist = FLT_MAX;
        int best_index = 0;

        const int mod_index = studydatas[study_trialindex].model_index;

        for (int i=0; i<ground_truths_gauges[mod_index].size(); ++i) {

            StudyData & d = ground_truths[mod_index][i];

            const float each_len = (vec2(pos.x(), pos.y()) - d.screen_pos).length2();

            if (each_len < best_dist) {
                best_dist = each_len;
                best_index = i;
            }

            //qDebug() << d.screen_pos << pos;
        }

        GaugeManipulator & g = ground_truths_gauges[mod_index][best_index];
        if (g.GetColourR() == 255 && g.GetColourG() == 255 && g.GetColourB() == 0) {
            g.SetColour(0, 255, 0);
        }
        else {
            g.SetColour(255, 255, 0);
        }

        qDebug() << "Model: " << meshnames[mod_index] << mod_index;
        for (int i=0; i<ground_truths[mod_index].size(); ++i) {
            GaugeManipulator & g = ground_truths_gauges[mod_index][i];
            qDebug() << (int)(g.GetColourR() == 255 && g.GetColourG() == 255 && g.GetColourB() == 0);
        }

    }

    mouse_left = false;
    mouse_middle = false;
    mouse_right = false;
    mouse_pos = pos;

    updateGL();

}

void GLWidget::keyPressEvent(QKeyEvent * event)
{

    switch (event->key()) {

    case Qt::Key_Shift:

        shift_key_held = true;

        break;

    case Qt::Key_Control:

        control_key_held = true;

        break;

    }

}

void GLWidget::keyReleaseEvent(QKeyEvent *event)
{ 

    switch (event->key()) {

    case Qt::Key_Shift:

        shift_key_held = false;

        break;

    case Qt::Key_Control:

        control_key_held = false;

        break;

    case Qt::Key_X:

        break;

    case Qt::Key_Y:

        break;

    case Qt::Key_Z:

        break;

    case Qt::Key_A:

        if (experimental_mode) {
            ComputeMedialAxisData();
        }

        break;

    case Qt::Key_B:

        if (experimental_mode) {
            //TODO: fix
            basRelief.ComputeBasReliefTransforms(turk_studydatas);
        }

        break;

    case Qt::Key_C:

        if (experimental_mode) {

            //ComputeHaltonSequences();
            //WriteHaltonSequences();

        }

        break;

    case Qt::Key_D:

        if (experimental_mode) {
            //ComputeAbstractionDistances();
            //ComputeCentroidDistances();
            //ComputeViewTrueNormAngle();
            ComputeCurvatureData();
        }

        break;

    case Qt::Key_E:

        if (experimental_mode) {
            ExportErrorGeometry();
        }

        break;

    case Qt::Key_F:

        if (experimental_mode) {
            ComputeAbstractionAngleDifferences();
        }

        break;

    case Qt::Key_G:

        if (experimental_mode) {
            draw_all_gauges = !draw_all_gauges;
        }

        break;

    case Qt::Key_H:

        if (experimental_mode) {
            DoAllHistogramAnalysis(turk_studydatas);
        }

        break;

    case Qt::Key_K:

        if (experimental_mode) {

            ComputeCurvatureData();

        }

        break;

    case Qt::Key_N:

        if (experimental_mode) {

            //AddCustomGauge(studydatas[study_trialindex].model_index);
            //WriteHaltonSequences();

        }

        break;

    case Qt::Key_P:

        if (experimental_mode) {

            ComputePValueTables();

        }

        break;

    case Qt::Key_T:

        if (experimental_mode) {

            ComputeLocalThickness();

        }

        break;

    case Qt::Key_Delete:

        break;

    case Qt::Key_Backspace:

        break;

    case Qt::Key_Left:       

        //if (coleStudyIndex > 0) {
        //    --coleStudyIndex;
        //}

        break;

    case Qt::Key_Right:

        //if (coleStudyIndex < 11) {
        //    ++coleStudyIndex;
        //}

        if (experimental_mode) {
            SetupNewTrial();
        }
        break;   

    case Qt::Key_Enter:
    case Qt::Key_Space:       

        if (studydatas[study_trialindex].num_gauge_manipulations > 0) {

            SetupNewTrial();

        }

        break;

    default:

        break;

    }

    updateGL();

}

int GLWidget::GetCurTrial()
{

    if (study_finished) {
        return GetNumTrials();
    }
    else {
        return study_trialindex;
    }

}

int GLWidget::GetNumTrials()
{

    return studydatas.size();

}

QList <int> GLWidget::GetRandomIndexes(int n)
{

    //randomize mesh indexes
    QList <int> meshinds;

    for (int i=0; i<n; ++i) {
        meshinds.push_back(i);
    }

    for (int i=0; i<n; ++i) {
        //swap randomly

        int randint = qrand() % n;

        int temp = meshinds[i];
        meshinds[i] = meshinds[randint];
        meshinds[randint] = temp;

    }

    return meshinds;

}

void GLWidget::SetupAllTrials()
{   

    //load trial_records file to set up trials (or makes it if it doesn't exist)
    if (!trial_records.Load("./records.txt")) {

        //make it if it doesn't exist
        trial_records.Create(meshnames.size(), num_total_sequence_gauges);
        trial_records.Save("./records.txt");

    }   

    //randomizes mesh order
    QList <int> meshinds = GetRandomIndexes(meshes.size());

    for (int i=0; i<meshinds.size(); ++i) {

        int mesh_index = meshinds[i];

        if (ground_truths[mesh_index].empty()) {
            continue;
        }

        //for each model, we pick a random set of 5 gauges (5 of the minimal set on record)
        QList <StudyData> eachmodellist;
        QList <int> rand_gauges;

        for (int j=0; j<study_numgaugespermodel; ++j) {

            const int rand_gauge = trial_records.ChooseMinimalRandom(mesh_index);
            rand_gauges.push_back(rand_gauge);

        }

        for (int k=0; k<2; ++k) {

            for (int j=0; j<study_numgaugespermodel; ++j) {

                //we make a random initial position for each trial
                GaugeManipulator each_g;
                each_g.SetNormal(meshstaticviews[mesh_index].ViewDir() * -1.0f);
                each_g.RotateNormalUD(meshstaticviews[mesh_index], qrand() % 181 - 90);
                each_g.RotateNormalLR(meshstaticviews[mesh_index], qrand() % 181 - 90);

                //index wraps around back to 0 when we hit 100, hence this modulo operation here
                const int cur_seq_index = rand_gauges[j];

                StudyData eachdata = ground_truths[mesh_index][cur_seq_index];

                eachdata.study_task = study_task;
                eachdata.abstraction = abstraction;
                eachdata.initial_normal = each_g.GetNormal();
                eachdata.guessed_normal = each_g.GetNormal();
                eachdata.num_cam_manipulations = 0;
                eachdata.num_gauge_manipulations = 0;

                eachmodellist.push_back(eachdata);

            }

        }

        //TODO: scramble the order before adding to studydata
        studydatas += eachmodellist;

    }

    qDebug() << "GLWidget::SetupAllTrials() - set up" << studydatas.size() << "trials";

}

void GLWidget::UpdateCamera() {

    makeCurrent();

    cam.SetWinWidth(width());
    cam.SetWinHeight(height());    
    cam.Update();
    cam.DrawGL_3DOrtho();

}

void GLWidget::SetupNewTrial()
{

    if (study_finished) {
        return;
    }

    //finish current trial if going on
    if (study_trialindex >= 0) {

        studydatas[study_trialindex].guessed_normal = gaugemanip.GetNormal();      
        studydatas[study_trialindex].trial_time_ms = study_trial_time.elapsed();       

    }

    //move on to next trial
    if (study_trialindex < studydatas.size() - 1) {       

        ++study_trialindex;       

        //start timer
        study_trial_time.start();

        //set up new gauge params
        gaugemanip.SetCentre(studydatas[study_trialindex].pos);
        gaugemanip.SetNormal(studydatas[study_trialindex].initial_normal);

        //move camera over that way
        bool animated_transition = false;

        //decide whether to animate
        if (study_task == ROTATE_VIEW) {
            if (study_trialindex == 0) {
                animated_transition = true;
            }
            else if (study_trialindex > 0) {

                const int model_ind1 = studydatas[study_trialindex-1].model_index;
                const int model_ind2 = studydatas[study_trialindex].model_index;

                //if model index changes, do camera animation
                animated_transition = (model_ind1 != model_ind2);
            }
        }

        //setup camera based on above decision
        if (!animated_transition ) {
            cam = meshstaticviews[studydatas[study_trialindex].model_index];
        }
        else {

            OrbitCamera start_cam = meshstaticviews[studydatas[study_trialindex].model_index];
            OrbitCamera end_cam = meshstaticviews[studydatas[study_trialindex].model_index];

            start_cam.SetPhi(start_cam.Phi() + 20.0f);
            start_cam.SetTheta(start_cam.Theta() - 90.0f);

            cam = start_cam;
            cam.SetInterpCamera(end_cam, 2000.0f);

        }

        if (load_and_show_study_data) {

            SetupTurkStudyDataGauges();

        }       

    }
    else {

        //QMessageBox::information(NULL, "Study Completed!", study_thanksmessage);
        //setEnabled(false);

        study_finished = true;
        study_successful = EvaluatePrecisionOfStudyData();

        if (experimental_mode) {
            qDebug() << "Note!  Auto-accepting results because experimental mode enabled.";
            study_successful = true;
        }

        if (study_successful) {

            CopyStudyDataToClipboard();

            //update # of successful runs
            trial_records.SetStudyRuns(trial_records.GetStudyRuns() + 1);
            //update the record file
            trial_records.Save("./records.txt");

        }

    }

    emit StartedNewTrial();
    emit DisableNextTrialButton();

}

void GLWidget::CopyStudyDataToClipboard()
{

    QClipboard * clipboard = QApplication::clipboard();   
    clipboard->setText(GetStudyData());

}

QString GLWidget::GetStudyData()
{

    QString studyData;

    studyData += "NewData ";

    studyData += turk_worker_id + " ";

    studyData += QDateTime::currentDateTime().toString("hh:mm:ss:zzz") + " ";

    studyData += QString::number(hashed_id) + " ";

    studyData += "run2 ";

    if (study_task == ROTATE) {
        studyData += "rotate ";
    }
    else if (study_task == ROTATE_VIEW) {
        studyData += "rotate_view ";
    }

    switch (abstraction) {
    case 0:
        studyData += "surface "; break;
    case 1:
        studyData += "mps "; break;
    case 2:
        studyData += "regxyz "; break;
    case 3:
        studyData += "radial "; break;
    case 4:
        studyData += "crdbrd "; break;
    case 5:
        studyData += "mod_mps "; break;

    }

    studyData += "\n";


    for (int i=0; i<studydatas.size(); ++i) {

        studyData += studydatas[i].ToString();

    }

    return studyData;

}

void GLWidget::LoadTurkStudyDataFromCSV(const QString & filename, TurkWorkerData & data)
{

    data.Clear();

    QFile file(filename);

    if (!file.open(QIODevice::ReadOnly | QIODevice::Text)) {
        qDebug() << "Turk study data CSV File " << filename << " can't be opened for reading";
        return;
    }

    QTextStream ifs(&file);

    QString eachline;

    while (!ifs.atEnd()) {

        eachline = ifs.readLine();

        parsenewdata:

        if (!eachline.contains("NewData", Qt::CaseSensitive) || !eachline.contains("Approved")) {
            continue;
        }

        eachline = eachline.right(eachline.length() - eachline.indexOf("NewData"));

        QStringList tokens = eachline.split(" ");

        QString id = tokens[1];
        QString hit_time = tokens[2];
        int condition = tokens[3].toInt();

        //if worker is outlier, skip
        if (data.IsWorkerOutlier(id)) {
            continue;
        }

        //add worker_id if it doesn't already exist
        if (!data.HasWorker(id)) {

            data.AddWorker(id, condition);

        }

        //skip this data if worker tried to cheat by copying data (will have the same timestamp)
        if (data.HasWorkerHit(id, hit_time)) {

            //qDebug() << "Worker" << worker_id << "copied data with timestamp" << worker_hit_time << "skipping";
            continue;

        }
        else if (data.GetWorkerNumHits(id) >= 10) {

            //only take the first 10 submissions
            continue;

        }
        else {

            //otherwise add this time
            data.AddWorkerHit(id, hit_time);

            //read the gauges
            for (int i=0; i<60; ++i) {

                eachline = ifs.readLine();
                eachline.remove("\"");

                StudyData d;

                if (d.FromString(eachline)) {

                    //add extra ground-truth data to this
                    d.pos = ground_truths[d.model_index][d.sequence_num].pos;
                    d.true_normal = ground_truths[d.model_index][d.sequence_num].true_normal;

                    d.task_error = AngleBetweenDeg(d.true_normal, d.guessed_normal);

                    data.AddWorkerHitGauge(id, hit_time, d);

                }
                else {

                    //remove worker's hit if not all 60 gauges there
                    data.RemoveWorkerHit(id, hit_time);

                    goto parsenewdata;

                }


            }

        }

    }   

    //DoAllHistogramAnalysis();

    //qDebug() << turk_studydatas_precise;
    //qDebug() << "Total data points: " << datas.size();

    file.close();

}

float GLWidget::GetMedian(QList <float> & samples)
{

    qSort(samples);

    const int samp_mid = samples.size() / 2;

    if (samples.size() % 2 == 0) {
        return (samples[samp_mid] + samples[samp_mid + 1]) * 0.5f;
    }
    else {
        return samples[samp_mid];
    }

}

float GLWidget::GetMean(QList <float> & samples)
{

    const int s = samples.size();
    float mean = 0.0f;

    for (int i=0; i<samples.size(); ++i) {

        mean += samples[i];

    }

    mean /= float(s);

    return mean;

}

float GLWidget::GetStandardDeviation(QList <float> & samples, float mean)
{

    const int s = samples.size();
    float sd = 0.0f;

    for (int i=0; i<s; ++i) {

        float diff = samples[i] - mean;
        sd += diff * diff;

    }

    sd /= float(s);
    sd = sqrtf(sd);

    return sd;

}

float GLWidget::GetNormalCurveArea(const float z)
{

    //approximation for cdf of normal curve - johnson et al (1994)
    /*
    const float y = 0.7988f*z*(1.0f+0.04417f*z*z);
    return expf(2.0f*y) / (1.0f + expf(2.0f*y));
    */

    //polya's (1945) approximation (error bound 0.003)
    //return 0.5f * (1.0f + sqrtf(1 - exp(-2.0f / M_PI * z * z)));

    //aludaat and alodat (2008) approximation (error bound 0.00197)
    return 0.5f + 0.5f * sqrt(1.0f - exp(-sqrtf(M_PI / 8.0f) * z * z));

}

float GLWidget::GetTwoSampleTwoTailZTestPValue(QList <float> & samples_i, QList <float> & samples_j)
{

    //compute means, std. devs, sample sizes for both datasets
    float mean_i = GetMean(samples_i);
    float mean_j = GetMean(samples_j);
    float stddev_i = GetStandardDeviation(samples_i, mean_i);
    float stddev_j = GetStandardDeviation(samples_j, mean_j);
    float n_i = samples_i.size();
    float n_j = samples_j.size();

    //compute p-value for dataset (i's distance to j)
    float z_value = fabsf((mean_i - mean_j) / sqrtf(stddev_i*stddev_i/n_i + stddev_j*stddev_j/n_j));

    //minusing by 0.5 this ignores the lower half area of the normal curve
    //(similar to tables in stats texts)
    float area_after_zero = GetNormalCurveArea(z_value) - 0.5f;

    //we do it two-tailed, goal to reject null hypothesis that mu_i = mu_j
    float p_value = 2 * (0.5f - area_after_zero);

    //qDebug() << data_i_samples;
    if (isnan(p_value)) {
        qDebug() << "GetTwoSampleTwoTailZTestPValue"
                 << "mean_i" << mean_i << "mean_j" << mean_j
                 << "std_i" << stddev_i << "std_j" << stddev_j
                 << "z_value" << z_value << "area_after_zero" << area_after_zero
                 << "pvalue" << p_value;
    }

    return p_value;

}

void GLWidget::SnapCameraToStaticView(int model_index) {

    cam = meshstaticviews[model_index];

}

void GLWidget::DoAllHistogramAnalysis(QList <StudyData> & studydatas)
{

    const float rot_task_hist_bucket_size = 5.0f;

    QVector <QList <float> > sample_errors;
    sample_errors.resize(meshes.size());

    if (study_task == 0) {

        //TODO: fix
        basRelief.ComputeBasReliefTransforms(studydatas);

        QList <float> angle_between_guesses_list;
        QList <float> angle_error_list;
        QList <float> angle_error_bas_list;

        Histogram angle_between_guesses;
        Histogram angle_error;
        Histogram angle_error_bas;

        for (int i=0; i<studydatas.size(); ++i) {

            //error with true normal test
            StudyData & d = studydatas[i];

            const float angle_error_sample = AngleBetweenDeg(d.guessed_normal, d.true_normal);
            const float bas_error_sample = AngleBetweenDeg(d.bas_corrected_normal, d.true_normal);

            studydatas[i].task_error = bas_error_sample;

            //qDebug() << d.bas_corrected_normal;

            angle_error.AddSample(angle_error_sample);
            angle_error_bas.AddSample(bas_error_sample);

            if (i < studydatas.size() - study_numgaugespermodel
                    && i % (study_numgaugespermodel * 2) < study_numgaugespermodel) {

                StudyData & d1 = studydatas[i];
                StudyData & d2 = studydatas[i + study_numgaugespermodel];

                const float angle_between_guesses_sample = AngleBetweenDeg(d1.guessed_normal, d2.guessed_normal);
                angle_between_guesses.AddSample(angle_between_guesses_sample);

            }

        }

        qDebug() << "histogram for angle_between_guesses_list (" << angle_between_guesses_list.size() << "samples)";
        //OutputHistogramData(angle_between_guesses_list, 0.0f, 180.0f, rot_task_hist_bucket_size);
        angle_between_guesses.SetRange(0.0f, 180.0f);
        angle_between_guesses.SetBucketSize(rot_task_hist_bucket_size);
        angle_between_guesses.OutputData();

        qDebug() << "histogram for angle_error_list (" << angle_error_list.size() << "samples)";
        //OutputHistogramData(angle_error_list, 0.0f, 180.0f, rot_task_hist_bucket_size);
        angle_error.SetRange(0.0f, 180.0f);
        angle_error.SetBucketSize(rot_task_hist_bucket_size);
        angle_error.OutputData();

        qDebug() << "histogram for angle_error_bas_list (" << angle_error_bas_list.size() << "samples)";
        //OutputHistogramData(angle_error_bas_list, 0.0f, 180.0f, rot_task_hist_bucket_size);
        angle_error_bas.SetRange(0.0f, 180.0f);
        angle_error_bas.SetBucketSize(rot_task_hist_bucket_size);
        angle_error_bas.OutputData();

    }

    //now do average errors for each model
    qDebug() << "Mesh, average rotation error, and bas-relief corrected average errors for models:";
    for (int m=0; m<meshes.size(); ++m) {

        for (int i=0; i<studydatas.size(); ++i) {

            StudyData & d = studydatas[i];

            if (d.model_index == m) {

                sample_errors[m].push_back(d.task_error);

            }

        }
    }

    for (int m=0; m<meshes.size(); ++m) {

        float mean = GetMean(sample_errors[m]);
        float median = GetMedian(sample_errors[m]);
        float std_dev = GetStandardDeviation(sample_errors[m], mean);
        float med_std_dev = GetStandardDeviation(sample_errors[m], median);

        qDebug() << meshnames[m] << mean << "(" << std_dev << ")" << median << "(" << med_std_dev << ")" << sample_errors[m].size();

    }


}

void GLWidget::SetupTurkStudyDataGauges()
{

    turk_studydata_gauges.clear();

    const int mod = studydatas[study_trialindex].model_index;
    const int cond = hashed_id;

    for (int i=0; i<ground_truths[mod].size(); ++i) {

        GaugeManipulator g;

        g.SetScale(gauge_scale);
        g.SetCentre(ground_truths[mod][i].pos);
        g.SetNormal(ground_truths[mod][i].true_normal);

        if (worker_data.GetDataPointNumSamples(mod, i, cond) > 0) {

            const float deg_apart = worker_data.GetDataPointAvgError(mod, i, cond);
            //int shade_val = (deg_apart > 60) ? 255 : int(255.0f * float(deg_apart) / 60.0f);
            //g.SetColour(shade_val, 255 - shade_val, 0); //green to red

            g.SetColourByError(deg_apart);

            turk_studydata_gauges.push_back(g);

        }
        else {

            //NOTE!  Not pushing back gauges with no samples!!

            g.SetColour(0, 0, 255); //green to red

        }       

    }

    /*
    //qDebug() << "turk study data size" << turk_studydatas.size();

    float avg_deg_apart = 0.0f;
    float avg_dist_apart = 0.0f;

    for (int i=0; i<turk_studydatas.size(); ++i) {

        if (studydatas[study_trialindex].model_index == turk_studydatas[i].model_index) {

            GaugeManipulator newgauge;
            int shade_val = 0;          

            //shade by angle between guessed/true normal
            int deg_apart = AngleBetweenDeg(turk_studydatas[i].guessed_normal, turk_studydatas[i].true_normal);

            avg_deg_apart += deg_apart;

            if (deg_apart > 60) {
                shade_val = 255;
            }
            else {
                shade_val = int(255.0f * float(deg_apart) / 60.0f);
            }

            newgauge.SetCentre(turk_studydatas[i].pos);
            newgauge.SetNormal(turk_studydatas[i].guessed_normal);

            float trial_time_sec = turk_studydatas[i].trial_time_ms * 0.001f;
            float time_scale = trial_time_sec * 0.01f * 0.1f;

            if (time_scale > 0.02f) {
                time_scale = 0.02f;
            }

            newgauge.SetScale(time_scale);
            newgauge.SetColour(shade_val, 255 - shade_val, 0); //green to red

            turk_studydata_gauges.push_back(newgauge);

        }

    }

    avg_deg_apart /= float(turk_studydata_gauges.size());
    avg_dist_apart /= float(turk_studydata_gauges.size());

    qDebug() << "avg_deg_apart" << avg_deg_apart << "avg_dist_apart" << avg_dist_apart;
    */

}

int GLWidget::GetRandomVisibleFace(const int mesh_index)
{

    const QVector <vec3> & verts = meshes[mesh_index].GetVerts();
    const QVector <QVector <int> > & faces = meshes[mesh_index].GetFaces();
    //const QVector <vec3> & facenorms = meshes[mesh_index].GetFaceNorms();

    //randomly select a face index for now
    int faceind = qrand() % faces.size();

    //calculate face centroid
    vec3 face_centroid = meshes[mesh_index].GetFaceCentroid(faceind);
    //qDebug() << "face centroid" << face_centroid;

    //project face centroid to screen
    cam = meshstaticviews[mesh_index];
    UpdateCamera();

    vec2 face_pt_2d = GLutils::ProjectPoint(face_centroid);
    //qDebug() << "face_pt_2d" << face_pt_2d;

    //construct ray to shoot back through face
    vec3 ray_p1;
    vec3 ray_p2;

    GLutils::UnProjectPoint(face_pt_2d, 0.1f, ray_p1);
    GLutils::UnProjectPoint(face_pt_2d, 0.9f, ray_p2);

    //qDebug() << "rayp1/2" << ray_p1 << ray_p2;

    ray1.push_back(ray_p1);
    ray2.push_back(ray_p2);

    //test intersection of ray with all triangles in the mesh, keep closest triangle as faceindex to use
    int closest_face_index = 0;
    float closest_face_dist = FLT_MAX;

    for (int i=0; i<faces.size(); ++i) {

        vec3 int_pt;
        if (GetRayTriIntersect(ray_p1, (ray_p2 - ray_p1).normalize(), verts[faces[i][0]], verts[faces[i][1]], verts[faces[i][2]], int_pt)) {

            //qDebug() << "ray hits face" << i;
            const float hit_dist = (cam.Eye() - int_pt).length();

            if (hit_dist < closest_face_dist) {

                closest_face_index = i;
                closest_face_dist = hit_dist;

            }

        }

    }

    //qDebug() << "returning closest index" << closest_face_index;

    return closest_face_index;

}

bool GLWidget::EvaluatePrecisionOfStudyData()
{      

    //test 1 - we require that 70% of the gauge pairs are within 30 degrees of each other
    //test 2 - standard deviation of all angle between gauge pairs is greater than 5 degrees

    int total_gauge_pairs = 0;
    int num_within_threshold = 0;

    const float thresh = 30.0f;
    float variance = 0.0f;

    for (int i=0; i<studydatas.size(); ++i) {

        if (i % (study_numgaugespermodel * 2) < study_numgaugespermodel) {

            ++total_gauge_pairs;

            StudyData & d1 = studydatas[i];
            StudyData & d2 = studydatas[i + study_numgaugespermodel];

            float delta = AngleBetweenDeg(d1.guessed_normal, d2.guessed_normal);

            variance += delta * delta; //add squared difference to variance

            if (delta < thresh) {
                ++num_within_threshold;
            }

        }

    }

    //normalize variance
    variance /= float(total_gauge_pairs);
    //std. dev. is just square root of variance
    const float std_dev = sqrtf(variance);

    const float ratio_within = float(num_within_threshold) / float(total_gauge_pairs);

    qDebug() << "Input evaluation: " << ratio_within << num_within_threshold << total_gauge_pairs << "std_dev: " << std_dev;

    return ratio_within >= 0.7f && std_dev >= 5.0f;

}

bool GLWidget::StudyFinished()
{
    return study_finished;
}

bool GLWidget::StudySuccessful()
{
    return study_successful;
}

void GLWidget::ComputeMedialAxisData()
{

    //what am i measuring?
    //1.  for surface only
    //2.  curvature at a gauge vs error at that gauge
    for (int c=0; c<10; ++c) {

        qDebug() << "MedAxisDist(X) Error(Y) for Condition " << c;

        for (int m=0; m<ground_truths.size(); ++m) {

            for (int s=0; s<ground_truths[m].size(); ++s) {

                if (worker_data.GetDataPointNumSamples(m, s, c) > 0) {
                    qDebug() << ground_truths[m][s].med_axis_dist << worker_data.GetDataPointAvgError(m ,s, c);
                }

            }
        }
    }

    /*
    qDebug() << "Performing medial axis analysis...";

    for(int mod=0; mod<ground_truths.size(); ++mod) {

        MedialAxis medial_axis;

        medial_axis.CreateFromSurface(meshes[mod].GetVerts(), meshes[mod].GetFaces(), 1.0f / 256.0f);
        medial_axis.WriteSurface(meshnames[mod] + "_voxel_surface.obj");
        medial_axis.WriteMedialAxis(meshnames[mod] + "_voxel_medial_axis.obj");

        Histogram medax_dists_hist;

        for (int seq=0; seq<ground_truths[mod].size(); ++seq) {

            const vec3 & pos = ground_truths[mod][seq].pos;
            medax_dists_hist.AddSample(medial_axis.AxisDistance(pos));

        }

        qDebug() << "Medial Axis data:";
        qDebug() << "X-medialaxisdist";

        qDebug() << medax_dists_hist.GetMean() << "(" << medax_dists_hist.GetStdDev() << ")" << meshnames[mod]; //express as % model diameter
        medax_dists_hist.OutputSamples();

    }
    */

}

void GLWidget::LoadCurvatures(const QString & filename, QList <float> & curvs)
{

    curvs.clear();

    QFile file(filename);

    if (!file.open(QIODevice::ReadOnly | QIODevice::Text)) {
        qDebug() << "File" << filename << " can't be opened for reading";
        return;
    }

    QTextStream ifs(&file);

    while (!ifs.atEnd()) {

        curvs.push_back(ifs.readLine().toFloat());

    }

    file.close();

}

float GLWidget::GetCurvatureForFace(const int model_index, const int face_index)
{

    const QVector <QVector <int> > & meshfaces = meshes[model_index].GetFaces();
    QVector <int> face = meshfaces[face_index];

    //first load up this curvature data
    QList <float> & curv1 = meshcurv1s[model_index]; //lists for each model of vertex curvature lists
    //QList <float> & curv2 = meshcurv2s[model_index];

    /*
    float gauss1 = curv1[face[0]] * curv2[face[0]];
    float gauss2 = curv1[face[1]] * curv2[face[1]];
    float gauss3 = curv1[face[2]] * curv2[face[2]];
    float avg_curv = (gauss1 + gauss2 + gauss3) / 3.0f;
    */

    /*
    float mean1 = (curv1[face[0]] + curv2[face[0]]) * 0.5f;
    float mean2 = (curv1[face[1]] + curv2[face[1]]) * 0.5f;
    float mean3 = (curv1[face[2]] + curv2[face[2]]) * 0.5f;
    float avg_curv = (mean1 + mean2 + mean3) / 3.0f;
    */

    float abs1 = fabsf(curv1[face[0]]);
    float abs2 = fabsf(curv1[face[0]]);
    float abs3 = fabsf(curv1[face[0]]);
    return qMax(abs1, qMax(abs2, abs3));

}

void GLWidget::ComputeCurvatureData()
{

    //let's find the optimal models to fit the curvature data
    /*
    for (int c=0; c<10; ++c) {

        float cur_best = 1.0f;
        const float delta = 0.001f;
        float cur_fit;
        float cur_fit_delta;
        int iterations = 0;

        do {

            Scatterplot cur_plot;
            Scatterplot cur_plot_delta;

            for (int m=0; m<ground_truths.size(); ++m) {

                for (int s=0; s<ground_truths[m].size(); ++s) {

                    if (worker_data.GetDataPointNumSamples(m, s, c) > 0) {

                        const float x = powf(fabsf(ground_truths[m][s].curv1), cur_best);
                        const float x_d = powf(fabsf(ground_truths[m][s].curv1), cur_best + delta);
                        const float y = worker_data.GetDataPointAvgError(m ,s, c);

                        cur_plot.AddSample(x, y);
                        cur_plot_delta.AddSample(x_d, y);

                    }

                }
            }

            //compute gradient of the residuals of fit
            cur_fit = cur_plot.GetRSquared();
            cur_fit_delta = cur_plot_delta.GetRSquared();

            //qDebug() << cur_best << cur_fit << cur_fit_delta;

            //plus because we want gradient ascension, that is, a bigger R-value
            cur_best += (cur_fit_delta - cur_fit);

            ++iterations;

        } while (fabsf(cur_fit_delta - cur_fit) > 0.00001f);

        qDebug() << "condition:" << c << "best:" << cur_best << "R^2:" << cur_fit << "iterations:" << iterations;

    }
    */

    //what am i measuring?
    //1.  for surface only
    //2.  curvature at a gauge vs error at that gauge

    for (int c=0; c<10; ++c) {

        qDebug() << "Curvature(X) Error(Y) for Condition " << c;

        for (int m=0; m<ground_truths.size(); ++m) {

            for (int s=0; s<ground_truths[m].size(); ++s) {

                if (worker_data.GetDataPointNumSamples(m, s, c) > 0) {
                    qDebug() << ground_truths[m][s].curv2 << worker_data.GetDataPointAvgError(m ,s, c);
                }

            }
        }
    }


    //load all the model curvature data
    /*
    meshcurv1s.resize(meshnames.size());
    meshcurv2s.resize(meshnames.size());

    for (int i=0; i<meshnames.size(); ++i) {

        LoadCurvatures(QString("./input_models/") + meshnames[i] + QString("/out_curv.txt"), meshcurv1s[i], meshcurv2s[i]);

    }

    for (int m=0; m<ground_truths.size(); ++m) {
        qDebug() << "Model" << meshnames[m] << "curv1 curv2";
        for (int g=0; g<ground_truths[m].size(); ++g) {

            //start with face
            const int face_ind = ground_truths[m][g].model_faceindex;

            //get verts for face
            QVector <int> vert_inds = meshes[m].GetFace(face_ind);

            //compute total dist to each vert of face
            float total_weight = 0.0f;
            for (int i=0; i<vert_inds.size(); ++i) {

                const float each_weight = 1.0f / (ground_truths[m][g].pos - meshes[m].GetVert(vert_inds[i])).length();

                if (each_weight > 0.0f) {
                    total_weight += 1.0f / (ground_truths[m][g].pos - meshes[m].GetVert(vert_inds[i])).length();
                }

            }

            //weighted sum of curv1 and curv2
            float blend_curv1 = 0.0f;
            float blend_curv2 = 0.0f;
            for (int i=0; i<vert_inds.size(); ++i) {

                const float each_weight = 1.0f / (ground_truths[m][g].pos - meshes[m].GetVert(vert_inds[i])).length();

                if (each_weight > 0.0f) {
                    blend_curv1 += meshcurv1s[m][vert_inds[i]] * each_weight / total_weight;
                    blend_curv2 += meshcurv2s[m][vert_inds[i]] * each_weight / total_weight;
                }

            }

            //calculate average "curvature" for the face
            ground_truths[m][g].curv1 = blend_curv1;
            ground_truths[m][g].curv2 = blend_curv2;
            ground_truths[m][g].curv_mean = (blend_curv1 + blend_curv2) * 0.5f;
            ground_truths[m][g].curv_gaussian = blend_curv1 * blend_curv2;

            qDebug() << ground_truths[m][g].curv1 << ground_truths[m][g].curv2;

        }
    }
    */

    /*
    qDebug() << "Performing curvature analysis...";

    QList <float> all_curvatures;
    QList <float> all_errors;

    QVector <QList <float> > curv_samples;
    QVector <QList <float> > error_samples;

    curv_samples.resize(meshes.size());
    error_samples.resize(meshes.size());

    for (int m=0; m<meshes.size(); ++m) {

        //const OFFFile & meshfile = meshes[studydatas[study_trialindex].model_index];        
        qDebug() << "Curvature data for" << meshnames[m];
        qDebug() << "X-curvature Y-error";

        for (int i=0; i<turk_studydatas.size(); ++i) {

            if (m == turk_studydatas[i].model_index) {

                //calculate average "curvature" for the face
                float curvature = GetCurvatureForFace(m, turk_studydatas[i].model_faceindex);

                //shade by dot prod of guessed/true normal
                int deg_apart = AngleBetweenDeg(turk_studydatas[i].guessed_normal, turk_studydatas[i].true_normal);

                all_curvatures.push_back(curvature);
                all_errors.push_back(deg_apart);

            }           

        }

    }

    //modify the visualization for the one we're looking at now
    turk_studydata_gauges.clear();
    for (int i=0; i<turk_studydatas.size(); ++i) {

        if (studydatas[study_trialindex].model_index == turk_studydatas[i].model_index) {

            GaugeManipulator newgauge;
            int shade_val = 0;          

            //shade by dot prod of guessed/true normal
            int deg_apart = AngleBetweenDeg(turk_studydatas[i].guessed_normal, turk_studydatas[i].true_normal);

            const float max_deg_apart = 30.0f;

            if (deg_apart > max_deg_apart) {
                shade_val = 255;
            }
            else {
                shade_val = int(255.0f * float(deg_apart) / max_deg_apart);
            }

            newgauge.SetCentre(turk_studydatas[i].pos);
            newgauge.SetNormal(turk_studydatas[i].guessed_normal);

            float gauge_scale = 0.02f * 0.01f * GetCurvatureForFace(turk_studydatas[i].model_index, turk_studydatas[i].model_faceindex);

            newgauge.SetScale(gauge_scale);
            newgauge.SetColour(shade_val, 255 - shade_val, 0); //green to red

            turk_studydata_gauges.push_back(newgauge);

        }

    }

    qDebug() << "Mean curvature, mean error";
    qDebug() << GetMean(all_curvatures) << GetMean(all_errors);
    //for (int i=0; i<all_curvatures.size(); ++i) {
    //    qDebug() << all_curvatures[i] << all_errors[i];
    //}

    //note - use this one if you want it per model
    //qDebug() << "Mean curvatures mean errors:";
    //for (int m=0; m<meshes.size(); ++m) {
    //    qDebug() << GetMean(curv_samples[m]) << GetMean(error_samples[m]);
    //}


    //OutputHistogramData(const QList <float> & list, 0.0f, 180.0f, 2.0f);
    */

    //what is good to show?  average error by medial axis buckets maybe?
    /*
    if (translation_study) {
        float tolerable_error = 0.02f;
        OutputHistogramLessThanData(all_curvatures, all_errors, tolerable_error, 0.0f, 200.0f, 1.0f);
        //OutputHistogramLessThanNormalizedData(all_medax_dists, all_errors, tolerable_error, 0.0f, 0.2f, 0.005f);
    }
    else {
        float tolerable_error = 30.0f;
        OutputHistogramLessThanData(all_curvatures, all_errors, tolerable_error, 0.0f, 200.0f, 1.0f);
        //OutputHistogramLessThanNormalizedData(all_medax_dists, all_errors, tolerable_error, 0.0f, 0.2f, 0.005f);
    }
    */

}

void GLWidget::VoxelListFromSurface(const OFFFile & model, const float voxel_size, QList <vec3> & voxel_pts)
{

    voxel_pts.clear();

    const QVector <vec3> & verts = model.GetVerts();
    const QVector <QVector <int> > & faces = model.GetFaces();

    MedialAxis med_axis;
    QList <VoxelPos> vox_pos_list = med_axis.GetSurfaceList(verts, faces, voxel_size);
    voxel_pts = med_axis.SurfaceListToVec(vox_pos_list);

}

int GLWidget::GetClosestPoint(const QList <vec3> & pts, const vec3 p)
{

    int closest_index = -1;

    float best_dist = FLT_MAX;

    for (int i=0; i<pts.size(); ++i) {

        float each_dist = (pts[i] - p).length2();

        if (each_dist < best_dist) {
            best_dist = each_dist;
            closest_index = i;
        }

    }

    return closest_index;

}

void GLWidget::ComputeAbstractionAngleDifferences()
{

    qDebug() << "AbsDiff(X) Error(Y) for Condition " << hashed_id;

    for (int m=0; m<ground_truths.size(); ++m) {

        for (int s=0; s<ground_truths[m].size(); ++s) {

            if (worker_data.GetDataPointNumSamples(m, s, hashed_id) > 0) {
                qDebug() << ground_truths[m][s].abstraction_angle_difference << worker_data.GetDataPointAvgError(m ,s, hashed_id);
            }

        }
    }

    /*
    qDebug() << "Surface-to-abstraction angle differences (NOTE!  Results specifically for condition" << hashed_id << "!)";

    for (int m=0; m<meshes.size(); ++m) {

        Histogram hist_abs_dists;

        //go through each face of mesh, perform ray-triangle test on inward facing normal direction
        const QVector <vec3> & verts = meshes[m].GetVerts();
        const QVector <QVector <int> > & faces = meshes[m].GetFaces();

        for (int i=0; i<ground_truths[m].size(); ++i) {

            //ray starts at face centroid
            //ray's direction is opposite face's normal
            const vec3 p0 = ground_truths[m][i].pos;

            float best_dist = FLT_MAX;
            int best_tri = 0;

            for (int j=0; j<faces.size(); ++j) {

                //we skip non triangle faces - the geometry is specifically crafted so that tris are planar faces, and quads are the extrusion faces
                //thus we can get the plane normals by only checking with triangles
                if (faces[j].size() != 3) {
                    continue;
                }

                const vec3 & p1 = verts[faces[j][0]];
                const vec3 & p2 = verts[faces[j][1]];
                const vec3 & p3 = verts[faces[j][2]];

                const float each_dist = PointTriangleDistance(p0, p1, p2, p3);

                if (each_dist < best_dist) {
                    best_dist = each_dist;
                    best_tri = j;
                }

            }

            const vec3 & p1 = verts[faces[best_tri][0]];
            const vec3 & p2 = verts[faces[best_tri][1]];
            const vec3 & p3 = verts[faces[best_tri][2]];

            //we take difference between the norm of this planar cut, and the true normal, to compare this difference with error
            const float angle_diff = AxisAngleDegrees(ground_truths[m][i].true_normal, (p2 - p1) ^ (p3 - p1));

            ground_truths[m][i].abstraction_angle_difference = angle_diff;

            hist_abs_dists.AddSample(angle_diff, QString("Gauge ") + QString::number(i));

            //if (i % 10 == 0) {
            //    qDebug() << "Processing gauge" << i << "/" << ground_truths[m].size()  << "for" << meshnames[m] << "abstraction_difference:" << ground_truths[m][i].abstraction_angle_difference;
            //}

        }

        QString filename = QString("./input_models/") + meshnames[m] + QString("/gauge_absdiff_");

        switch (abstraction) {
        case 1:
            filename += QString("mps.txt");
            break;
        case 2:
            filename += QString("xyz.txt");
            break;
        case 3:
            filename += QString("radial.txt");
            break;
        case 4:
            filename += QString("crdbrd.txt");
            break;
        }


        hist_abs_dists.OutputSamplesToFile(filename);

    }
    */

}

//analysis - view-truenorm angle
void GLWidget::ComputeViewTrueNormAngle()
{

    qDebug() << "ViewNorm(X) Error(Y) for Condition " << hashed_id;

    for (int m=0; m<ground_truths.size(); ++m) {

        for (int s=0; s<ground_truths[m].size(); ++s) {

            if (worker_data.GetDataPointNumSamples(m, s, hashed_id) > 0) {
                qDebug() << ground_truths[m][s].viewnorm_angle << worker_data.GetDataPointAvgError(m, s, hashed_id);
            }

        }
    }

    //this does the actual value computation
    /*
    qDebug() << "View-norm angles";

    for (int m=0; m<meshes.size(); ++m) {

        Histogram hist_viewnorm_angles;

        const vec3 view_dir = -meshstaticviews[m].ViewDir().normalize();

        for (int i=0; i<ground_truths[m].size(); ++i) {

            const float viewnorm_angle = acosf(ground_truths[m][i].true_normal * view_dir) * RAD_TO_DEG;
            ground_truths[m][i].viewnorm_angle = viewnorm_angle;

            hist_viewnorm_angles.AddSample(viewnorm_angle, QString("Gauge ") + QString::number(i));

        }

        qDebug() << "Viewnorm angles for model" << meshnames[m] << "--------------------------------";
        hist_viewnorm_angles.OutputSamples();

    }
    */

}

void GLWidget::ComputeCentroidDistances()
{

    qDebug() << "CentDist(X) Error(Y) for Condition " << hashed_id;

    for (int m=0; m<ground_truths.size(); ++m) {

        for (int s=0; s<ground_truths[m].size(); ++s) {

            if (worker_data.GetDataPointNumSamples(m, s, hashed_id) > 0) {
                qDebug() << ground_truths[m][s].centroid_distance << worker_data.GetDataPointAvgError(m, s, hashed_id);
            }

        }
    }
    //this does the actual value computation
    /*
    qDebug() << "Surface-to-centroid distances";

    for (int m=0; m<meshes.size(); ++m) {

        Histogram hist_cent_dists;

        //go through each face of mesh, perform ray-triangle test on inward facing normal direction
        const QVector <vec3> & verts = meshes[m].GetVerts();
        //const QVector <QVector <int> > & faces = meshes[m].GetFaces();

        //compute centroid position based on vertices (use bounding box midpoint)
        vec3 lo(FLT_MAX, FLT_MAX, FLT_MAX);
        vec3 hi(-FLT_MAX, -FLT_MAX, -FLT_MAX);
        for (int i=0; i<verts.size(); ++i) {
            for (int j=0; j<3; ++j) {
                lo[j] = qMin(lo[j], verts[i][j]);
                hi[j] = qMax(hi[j], verts[i][j]);
            }
        }
        const vec3 centroid = lo + (hi - lo) * 0.5f;

        for (int i=0; i<ground_truths[m].size(); ++i) {

            const float cent_dist = (ground_truths[m][i].pos - centroid).length();
            ground_truths[m][i].centroid_distance = cent_dist;

            hist_cent_dists.AddSample(cent_dist, QString("Gauge ") + QString::number(i));

        }

        qDebug() << "Centroid distances for model" << meshnames[m] << "--------------------------------";
        hist_cent_dists.OutputSamples();

    }
    */


}

void GLWidget::ComputeAbstractionDistances()
{

    qDebug() << "AbsDist(X) Error(Y) for Condition " << hashed_id;

    for (int m=0; m<ground_truths.size(); ++m) {

        for (int s=0; s<ground_truths[m].size(); ++s) {

            if (worker_data.GetDataPointNumSamples(m, s, hashed_id) > 0) {
                qDebug() << ground_truths[m][s].abstraction_distance << worker_data.GetDataPointAvgError(m ,s, hashed_id);
            }

        }
    }

    //this does the actual value computation
    /*
    qDebug() << "Surface-to-abstraction distances (NOTE!  Results specifically for condition" << hashed_id << "!)";

    for (int m=0; m<meshes.size(); ++m) {

        Histogram hist_abs_dists;

        //go through each face of mesh, perform ray-triangle test on inward facing normal direction
        const QVector <vec3> & verts = meshes[m].GetVerts();
        const QVector <QVector <int> > & faces = meshes[m].GetFaces();

        for (int i=0; i<ground_truths[m].size(); ++i) {

            //ray starts at face centroid
            //ray's direction is opposite face's normal
            const vec3 p0 = ground_truths[m][i].pos;

            float best_dist = FLT_MAX;

            for (int j=0; j<faces.size(); ++j) {

                //face may not be tri - go through whole fan
                for (int k=2; k<faces[j].size(); ++k) {

                    const vec3 p1 = verts[faces[j][0]];
                    const vec3 p2 = verts[faces[j][k-1]];
                    const vec3 p3 = verts[faces[j][k]];

                    best_dist = qMin(best_dist, PointTriangleDistance(p0, p1, p2, p3));

                }

            }

            ground_truths[m][i].abstraction_distance = best_dist;
            hist_abs_dists.AddSample(best_dist, QString("Gauge ") + QString::number(i));

            if (i % 10 == 0) {
                qDebug() << "Processing gauge" << i << "/" << ground_truths[m].size()  << "for" << meshnames[m] << "abstraction_distance:" << ground_truths[m][i].abstraction_distance;
            }

        }

        qDebug() << "Abstraction distances for model" << meshnames[m] << "--------------------------------";
        hist_abs_dists.OutputSamples();

    }
    */

}

void GLWidget::ExportErrorGeometry()
{

    for (int m=0; m<meshes.size(); ++m) {

        OFFFile figure_geom_big_pos;
        OFFFile figure_geom_medium_pos;
        OFFFile figure_geom_small_pos;
        OFFFile figure_geom_smaller_pos;
        OFFFile figure_geom_none;
        OFFFile figure_geom_small_neg;
        OFFFile figure_geom_big_neg;

        int n_cubes = 0;

        for (int i=0; i<turk_studydatas.size(); ++i) {                      

            //float cube_size = 0.015f;
            float disc_radius = 0.01f;

            if (i < turk_studydatas.size() - study_numgaugespermodel
                    && i % (study_numgaugespermodel * 2) < study_numgaugespermodel) {

                StudyData & d1 = turk_studydatas[i];
                StudyData & d2 = turk_studydatas[i + study_numgaugespermodel];

                if (d1.model_index != m || d2.model_index != m) {
                    continue;
                }

                //give it a colour (red for out from normal, blue for inside normal)
                float angle_error = (AngleBetweenDeg(d1.guessed_normal, d1.true_normal) + AngleBetweenDeg(d1.guessed_normal, d1.true_normal)) * 0.5f;

                if (angle_error > 60.0f) {
                    //figure_geom_big_pos.AddCube(d.pos + vec3(-cube_size/2.0f, -cube_size/2.0f, -cube_size/2.0f), cube_size);
                    figure_geom_big_pos.AddDisc(d1.pos, d1.true_normal, disc_radius);
                }
                else if (angle_error > 45.0f) {
                    //figure_geom_small_pos.AddCube(d.pos + vec3(-cube_size/2.0f, -cube_size/2.0f, -cube_size/2.0f), cube_size);
                    figure_geom_medium_pos.AddDisc(d1.pos, d1.true_normal, disc_radius);
                }
                else if (angle_error > 30.0f) {
                    //figure_geom_small_pos.AddCube(d.pos + vec3(-cube_size/2.0f, -cube_size/2.0f, -cube_size/2.0f), cube_size);
                    figure_geom_small_pos.AddDisc(d1.pos, d1.true_normal, disc_radius);
                }
                else if (angle_error > 15.0f) {
                    //figure_geom_smaller_pos.AddCube(d.pos + vec3(-cube_size/2.0f, -cube_size/2.0f, -cube_size/2.0f), cube_size);
                    figure_geom_smaller_pos.AddDisc(d1.pos, d1.true_normal, disc_radius);
                }
                else {
                    //figure_geom_none.AddCube(d.pos + vec3(-cube_size/2.0f, -cube_size/2.0f, -cube_size/2.0f), cube_size);
                    figure_geom_none.AddDisc(d1.pos, d1.true_normal, disc_radius);
                }

                ++n_cubes;

            }

        }

        qDebug() << n_cubes << "cubes";
        if (study_task == ROTATE) {
            figure_geom_big_pos.Write_OBJ(QString("Ro_error_") + meshnames[m] + "_big_pos.obj");
            figure_geom_medium_pos.Write_OBJ(QString("Ro_error_") + meshnames[m] + "_medium_pos.obj");
            figure_geom_small_pos.Write_OBJ(QString("Ro_error_") + meshnames[m] + "_small_pos.obj");
            figure_geom_smaller_pos.Write_OBJ(QString("Ro_error_") + meshnames[m] + "_smaller_pos.obj");
            figure_geom_none.Write_OBJ(QString("Ro_error_") + meshnames[m] + "_none.obj");
        }
        else if (study_task == ROTATE_VIEW) {
            figure_geom_big_pos.Write_OBJ(QString("RoVi_error_") + meshnames[m] + "_big_pos.obj");
            figure_geom_big_neg.Write_OBJ(QString("RoVi_error_") + meshnames[m] + "_big_neg.obj");
            figure_geom_small_pos.Write_OBJ(QString("RoVi_error_") + meshnames[m] + "_small_pos.obj");
            figure_geom_small_neg.Write_OBJ(QString("RoVi_error_") + meshnames[m] + "_small_neg.obj");
            figure_geom_none.Write_OBJ(QString("RoVi_error_") + meshnames[m] + "_none.obj");
        }       

    }

}

void GLWidget::ComputeLocalThickness()
{

    //this stuff outputs scatterplots for thickness vs error for each condition
    for (int c=0; c<10; ++c) {

        qDebug() << "LocalThickness(X) Error(Y) for Condition " << c;

        for (int m=0; m<ground_truths.size(); ++m) {

            Histogram hist_thick;

            for (int s=0; s<ground_truths[m].size(); ++s) {

                if (worker_data.GetDataPointNumSamples(m, s, c) > 0) {
                    //qDebug() << ground_truths[m][s].local_thickness << worker_data.GetDataPointAvgError(m ,s, c);

                    if (c == 0) {
                        hist_thick.AddSample(ground_truths[m][s].local_thickness);
                    }
                }
            }

            if (c == 0) {
                qDebug() << meshnames[m] << hist_thick.GetMean() << hist_thick.GetStdDev() << hist_thick.GetMin() << hist_thick.GetMax();
            }
        }
    }

    //this stuff actually does the ray-tracing to get thickness estimates at each gauge
    /*
    for (int m=0; m<meshes.size(); ++m) {

        //go through each face of mesh, perform ray-triangle test on inward facing normal direction
        const QVector <vec3> & verts = meshes[m].GetVerts();
        const QVector <QVector <int> > & faces = meshes[m].GetFaces();

        for (int i=0; i<ground_truths[m].size(); ++i) {

            //ray starts at face centroid
            //ray's direction is opposite face's normal
            const vec3 ray_p = ground_truths[m][i].pos;
            const vec3 ray_d = ground_truths[m][i].true_normal * (-1.0f);

            float best_hit_dist = FLT_MAX;

            for (int j=0; j<faces.size(); ++j) {

                //skip face that this gauge lies on
                if (ground_truths[m][i].model_faceindex == j) {
                    continue;
                }

                //face may not be tri - go through whole fan
                for (int k=2; k<faces[j].size(); ++k) {

                    const vec3 p1 = verts[faces[j][0]];
                    const vec3 p2 = verts[faces[j][k-1]];
                    const vec3 p3 = verts[faces[j][k]];

                    vec3 ipt;
                    const bool hit = GetRayTriIntersect(ray_p, ray_d, p1, p2, p3, ipt);

                    if (hit) {

                        const float dist = (ipt - ray_p).length2();

                        if (dist < best_hit_dist && dist > 0.00001f) {
                            best_hit_dist = dist;

                            //qDebug() << "HIT! mesh" << meshnames[m] << "ray" << i << "hits face" << j << "with dist" << dist;
                            //qDebug() << "  triverts" << p1 << p2 << p3;
                            //qDebug() << "  ray" << ray_p << ray_d;
                        }

                    }

                }

            }

            ground_truths[m][i].local_thickness = best_hit_dist;

            if (i % 10 == 0) {
                qDebug() << "Processing gauge" << i << "/" << ground_truths[m].size()  << "for" << meshnames[m] << "thickness:" << ground_truths[m][i].local_thickness;
            }

        }       

    }

    //output it all
    for (int m=0; m<ground_truths.size(); ++m) {
        qDebug() << "Local thicknesses for each gauge for model" << meshnames[m];

        for (int i=0; i<ground_truths[m].size(); ++i) {
            qDebug() << ground_truths[m][i].local_thickness;
        }
    }
    */

}

void GLWidget::ComputePValueTables()
{

    /*
    const int nDatas = turk_studydata_filenames.size();

    //load all turk studydatas for task
    QVector <QList <StudyData> > all_task_studydata;
    all_task_studydata.resize(nDatas);

    for (int i=0; i<nDatas; ++i) {
        LoadTurkStudyDataFromCSV(turk_studydata_filenames[i], all_task_studydata[i]);
    }

    //process all errors for it
    for (int i=0; i<nDatas; ++i) {
        DoAllHistogramAnalysis(all_task_studydata[i]);
    }

    //write all to file
    QString filename = QString("all_errors_") + QString::number(study_task) + QString(".txt");
    QFile file(filename);
    if (!file.open(QIODevice::WriteOnly | QIODevice::Text)) {
        qDebug() << "cant write" << filename;
        return;
    }
    qDebug() << "Writing" << filename << "...";

    QTextStream ofs(&file);

    for (int i=0; i<all_task_studydata.size(); ++i) {
        for (int j=0; j<all_task_studydata[i].size(); ++j) {

            StudyData & d = all_task_studydata[i][j];
            if (d.IncludeData(study_task)) {
                //task, model, abstraction (NOTE 0 is baseline, 1 is mps, etc...), error
                ofs << study_task << " " << d.model_index << " " << i << " " << d.DataError(study_task) << "\n";
            }

        }
    }

    file.close();
    */

    //a giant loader for all our simplified data
    QString filename = QString("all_errors.txt");
    QFile file(filename);
    if (!file.open(QIODevice::ReadOnly | QIODevice::Text)) {
        qDebug() << "cant read" << filename;
        return;
    }
    qDebug() << "Reading" << filename << "...";

    QTextStream ifs(&file);

    QList <StudyData> all_data_list;

    while (!ifs.atEnd()) {

        QStringList eachline = ifs.readLine().split(" ");
        if (eachline.size() == 4) {
            StudyData d;
            d.study_task = eachline[0].toInt() == 0 ? ROTATE : ROTATE_VIEW;
            d.model_index = eachline[1].toInt();
            d.abstraction = eachline[2].toInt();
            d.task_error = eachline[3].toFloat();
            all_data_list.push_back(d);
        }

    }

    file.close();


    //next, go through each pair and compute p-values... do this ONLY for each model though
    int nAbstractions = 6; //0 to 5 (note 0 is baseline, 5 is crdbrd)

    for (int t=0; t<3; ++t) { //task t

        for (int m=0; m<meshes.size(); ++m) { //model m

            QVector <QList <float> > all_p_values;
            all_p_values.resize(nAbstractions); //there are 6 abstractions

            for (int i=0; i<nAbstractions; ++i) { //abstraction i
                for (int j=0; j<nAbstractions; ++j) { //abstraction j

                    //take samples for this model from both datasets
                    QList <float> data_i_samples;
                    QList <float> data_j_samples;

                    //populate data_i_samples for error
                    for (int k=0; k<all_data_list.size(); ++k) {

                        StudyData & d = all_data_list[k];

                        if (d.study_task == t && d.model_index == m && d.abstraction == i) {
                            data_i_samples.push_back(d.task_error);
                        }

                    }

                    //populate data_j_samples for error
                    for (int k=0; k<all_data_list.size(); ++k) {

                        StudyData & d = all_data_list[k];

                        if (d.study_task == t && d.model_index == m && d.abstraction == j) {
                            data_j_samples.push_back(d.task_error);
                        }

                    }

                    float p_value = GetTwoSampleTwoTailZTestPValue(data_i_samples, data_j_samples);

                    all_p_values[i].push_back(p_value);

                }
            }

            //output
            qDebug();
            //qDebug() << GetNormalCurveArea(2.0f);
            qDebug() << "Table of P-values for study-task" << t << "model" << meshnames[m];
            qDebug();
            for (int i=0; i<all_p_values.size(); ++i) {

                QString eachline;

                for (int j=0; j<all_p_values[i].size(); ++j) {
                    eachline += QString::number(all_p_values[i][j]) + " ";
                }

                qDebug() << qPrintable(eachline);
            }

            QString filename = QString("task") + QString::number(t) + QString("model") + QString::number(m) + QString(".html");
            QFile file(filename);
             if (file.open(QIODevice::WriteOnly | QIODevice::Text)) {
                 QTextStream os(&file);
                 OutputHTMLTable(os, all_p_values);
                 file.close();
             }
             else {
                 qDebug() << "Cant write" << filename;
             }


        }

    }

    //outputting per-model-per-abstraction histograms
    StudyTask target_task = ROTATE;
    int target_model = 5;

    for (int a=0; a<6; ++a) {

        Histogram target_hist;

        for (int i=0; i<all_data_list.size(); ++i) {

            StudyData & d = all_data_list[i];

            if (d.study_task == target_task && d.model_index == target_model && d.abstraction == a) {
                target_hist.AddSample(d.task_error);
            }

        }

        qDebug() << "Histogram for" << target_hist.GetNumSamples() << "samples:" << target_task << target_model << a;

        target_hist.SetRange(0.0f, 180.0f);
        target_hist.SetBucketSize(5.0f);
        target_hist.OutputData();

    }


}


void GLWidget::OutputHTMLTable(QTextStream & os, const QVector <QList <float> > all_values) {

    QList <vec3> heading_colours;
    heading_colours.push_back(vec3(0, 69, 134));
    heading_colours.push_back(vec3(255, 66, 14));
    heading_colours.push_back(vec3(255, 211, 32));
    heading_colours.push_back(vec3(87, 157, 28));
    heading_colours.push_back(vec3(126, 0, 33));
    heading_colours.push_back(vec3(131, 202, 255));
    heading_colours.push_back(vec3(49, 64, 4));

    os.setFieldWidth(2);
    os.setPadChar(QChar('0'));

    os << "<html>\n";
    os << "<head>\n";
    os << "</head>\n";
    os << "<body>\n";

    os << "<table border=0 cellpadding=5 cellspacing=0>\n";
    os << "<tr height=40px>\n";
    os << "<th width=40px></th>\n";

    for (int i=0; i<all_values.size(); ++i) {

        hex(os);
        os << "<th width=40px bgcolor=#"
           << int(heading_colours[i][0])
                      << int(heading_colours[i][1])
                      << int(heading_colours[i][2]) << "></th>\n";
    }

    os << "</tr>\n";

    for (int i=0; i<all_values.size(); ++i) {

        os << "<tr height=40px>\n";

        hex(os);
        os << "<td bgcolor=#" << int(heading_colours[i][0])
           << int(heading_colours[i][1])
           << int(heading_colours[i][2]) << "></td>\n";

        for (int j=0; j<all_values[i].size(); ++j) {

            float val_scaled = ( all_values[i][j] / 0.05f );

            val_scaled = qMax(val_scaled, 0.0f);
            val_scaled = qMin(val_scaled, 1.0f);

            int hexValue = int(val_scaled * 255.0f);

            /*
            os << "<td bgcolor=#FF" << std::hex << std::uppercase << 255-hexValue
                    << std::hex << std::uppercase << 255-hexValue
                    << ">" << std::dec << fData1[i][j] << " / " << fData2[i][j] << "</td>" << std::endl;
                    */
            hex(os);
            os << "<td bgcolor=#" << 255-hexValue << 255-hexValue << 255-hexValue << "></td>\n";

        }

        os << "</tr>\n";

    }

    os << "</table>\n";

    os << "</body>\n";
    os << "</html>\n";

}

float GLWidget::GetHaltonSequence(const int index, const int base)
{

    float result = 0.0f;
    float f = 1.0f / float(base);
    int i = index + 1;

    while (i > 0) {

        result += f * (i % base);
        i /= base;
        f /= base;

    }

    return result;

}

bool GLWidget::GetGaugeForPoint(const QPoint & point, const int mesh_index, StudyData & d)
{

    cam = meshstaticviews[mesh_index];
    UpdateCamera();

    vec2 point_2d(point.x(), point.y());

    int face_index;
    vec3 int_pt;

    if (ClosestFaceAlongRay(mesh_index, point_2d, face_index, int_pt)) {

        d.model_index = mesh_index;
        d.screen_pos = point_2d;
        d.model_faceindex = face_index;
        d.pos = int_pt;
        d.true_normal = meshes[mesh_index].GetFaceNorms()[face_index];

        return true;

    }
    else {

        return false;

    }

}

void GLWidget::ComputeHaltonSequences()
{

    ground_truths.clear();
    ground_truths.resize(meshnames.size());

    ground_truths_gauges.clear();
    ground_truths_gauges.resize(meshnames.size());

    for (int i=0; i<meshes.size(); ++i) {

        QList <StudyData> halton_pts;
        int halton_seq = 0;      

        while (ground_truths[i].size() < num_halton_sequence_gauges) {

            //get a 2d point
            ++halton_seq;
            const float x_val = float(GetHaltonSequence(halton_seq, 2)) * width();
            const float y_val = float(GetHaltonSequence(halton_seq, 3)) * height();
            QPoint p_2d(x_val, y_val);

            StudyData d;
            if (GetGaugeForPoint(p_2d, i, d)) {

                ground_truths[i].push_back(d);

            }

        }

    }

}

void GLWidget::WriteHaltonSequences()
{

    for (int i=0; i<meshes.size(); ++i) {

        //write out halton sequence to a file
        //write data out to file
        QString filename = QString("./input_models/") + meshnames[i] + QString("/halton_sequence.txt");
        QFile file(filename);

        if (!file.open(QIODevice::WriteOnly | QIODevice::Text)) {
            qDebug() << "File " << filename << " can't be written";
            return;
        }

        qDebug() << "Writing" << filename << "...";

        QTextStream ofs(&file);

        for (int j=0; j<ground_truths[i].size(); ++j) {

            const StudyData & d = ground_truths[i][j];

            ofs << j << " "
                << d.screen_pos[0] << " "
                << d.screen_pos[1] << " "
                << d.model_faceindex << " "
                << d.pos[0] << " "
                << d.pos[1] << " "
                << d.pos[2] << " "
                << d.true_normal[0] << " "
                << d.true_normal[1] << " "
                << d.true_normal[2] << "\n";

        }

        file.close();

    }

}

bool GLWidget::ClosestFaceAlongRay(const int mesh_index, const vec2 & screen_pt, int & face_index, vec3 & int_pt)
{

    const QVector <vec3> & verts = meshes[mesh_index].GetVerts();
    const QVector <QVector <int> > & faces = meshes[mesh_index].GetFaces();

    //construct ray to shoot back through face
    vec3 ray_p1;
    vec3 ray_p2;

    GLutils::UnProjectPoint(screen_pt, 0.1f, ray_p1);
    GLutils::UnProjectPoint(screen_pt, 0.9f, ray_p2);

    ray1.push_back(ray_p1);
    ray2.push_back(ray_p2);

    //test intersection of ray with all triangles in the mesh, keep closest triangle as faceindex to use
    face_index = -1;
    float face_dist = FLT_MAX;

    for (int i=0; i<faces.size(); ++i) {

        vec3 each_int_pt;

        if (GetRayTriIntersect(ray_p1, (ray_p2 - ray_p1).normalize(), verts[faces[i][0]], verts[faces[i][1]], verts[faces[i][2]], each_int_pt)) {

            //qDebug() << "ray hits face" << i;
            const float hit_dist = (cam.Eye() - each_int_pt).length2();

            if (hit_dist < face_dist) {

                face_index = i;
                face_dist = hit_dist;
                int_pt = each_int_pt;

            }

        }

    }

    return face_index >= 0;

}

void GLWidget::LoadHaltonSequences()
{

    ground_truths.clear();
    ground_truths.resize(meshnames.size());

    ground_truths_gauges.clear();
    ground_truths_gauges.resize(meshnames.size());

    for (int i=0; i<meshnames.size(); ++i) {

        QString filename = QString("./input_models/") + meshnames[i] + QString("/halton_sequence.txt");
        QFile file(filename);

        if (!file.open(QIODevice::ReadOnly | QIODevice::Text)) {
            qDebug() << "File " << filename << " can't be read";
            continue;
        }

        qDebug() << "Reading" << filename << "...";

        QTextStream ifs(&file);

        while (!ifs.atEnd()) {

            QStringList eachline = ifs.readLine().split(" ");

            StudyData d;

            d.model_index = i;
            d.sequence_num = eachline[0].toInt();
            d.screen_pos[0] = eachline[1].toFloat();
            d.screen_pos[1] = eachline[2].toFloat();
            d.model_faceindex = eachline[3].toInt();
            d.pos[0] = eachline[4].toFloat();
            d.pos[1] = eachline[5].toFloat();
            d.pos[2] = eachline[6].toFloat();
            d.true_normal[0] = eachline[7].toFloat();
            d.true_normal[1] = eachline[8].toFloat();
            d.true_normal[2] = eachline[9].toFloat();

            ground_truths[i].push_back(d);

            GaugeManipulator g;

            g.SetCentre(d.pos);
            g.SetNormal(d.true_normal);
            g.SetScale(gauge_scale);
            if (ground_truths[i].size() > num_halton_sequence_gauges) {
                g.SetColour(0, 0, 255);
            }

            ground_truths_gauges[i].push_back(g);

        }

        file.close();

        //load "flatness" info for gauge
        if (experimental_mode) {

            QString filename = QString("./input_models/") + meshnames[i] + QString("/gauge_flatness.txt");
            QFile file(filename);

            if (!file.open(QIODevice::ReadOnly | QIODevice::Text)) {
                qDebug() << "File " << filename << " can't be read";
                continue;
            }

            qDebug() << "Reading" << filename << "...";

            QTextStream ifs(&file);

            for (int j=0; j<ground_truths[i].size(); ++j) {

                QString eachline = ifs.readLine();

                ground_truths[i][j].flatness = (eachline.toInt() == 1);

                if (ground_truths[i][j].flatness) {
                    ground_truths_gauges[i][j].SetColour(255, 255, 0);
                }

            }

            file.close();

        }

        //load "curvature" info for gauge
        if (experimental_mode) {

            QString filename1 = QString("./input_models/") + meshnames[i] + QString("/gauge_curv1.txt");
            QString filename2 = QString("./input_models/") + meshnames[i] + QString("/gauge_curv2.txt");

            QList <float> curv1s;
            QList <float> curv2s;

            LoadCurvatures(filename1, curv1s);
            LoadCurvatures(filename2, curv2s);

            for (int j=0; j<ground_truths[i].size(); ++j) {

                ground_truths[i][j].curv1 = curv1s[j];
                ground_truths[i][j].curv2 = curv2s[j];
                ground_truths[i][j].curv_gaussian = ground_truths[i][j].curv1 * ground_truths[i][j].curv2;
                ground_truths[i][j].curv_mean = (ground_truths[i][j].curv1 + ground_truths[i][j].curv2) * 0.5f;

                const GLubyte curv_col = qMin(fabsf(ground_truths[i][j].curv1) * 10.0f, 255.0f);
                if (ground_truths[i][j].curv1 > 0.0f) {
                    ground_truths_gauges[i][j].SetColour(255, 255 - curv_col, 255 - curv_col);
                }
                else {
                    ground_truths_gauges[i][j].SetColour(255 - curv_col, 255 - curv_col, 255);
                }

            }

        }

        //load "medial axis distance" info for gauge
        if (experimental_mode) {

            QString filename = QString("./input_models/") + meshnames[i] + QString("/gauge_med_axis_dist.txt");
            QFile file(filename);

            if (!file.open(QIODevice::ReadOnly | QIODevice::Text)) {
                qDebug() << "File " << filename << " can't be read";
                continue;
            }

            qDebug() << "Reading" << filename << "...";

            QTextStream ifs(&file);

            for (int j=0; j<ground_truths[i].size(); ++j) {

                QString eachline = ifs.readLine();

                ground_truths[i][j].med_axis_dist = eachline.toFloat();

                //colour gauges based on the distance
                const GLubyte curv_col = qMin(fabsf(ground_truths[i][j].med_axis_dist) * 10.0f * 255.0f, 255.0f);
                ground_truths_gauges[i][j].SetColour(255, 255 - curv_col, 255 - curv_col);

            }

            file.close();

        }

        //load "local thickness" info for gauge
        if (experimental_mode) {

            QString filename = QString("./input_models/") + meshnames[i] + QString("/gauge_thickness.txt");
            QFile file(filename);

            if (!file.open(QIODevice::ReadOnly | QIODevice::Text)) {
                qDebug() << "File " << filename << " can't be read";
                continue;
            }

            qDebug() << "Reading" << filename << "...";

            QTextStream ifs(&file);

            for (int j=0; j<ground_truths[i].size(); ++j) {

                QString eachline = ifs.readLine();

                ground_truths[i][j].local_thickness = eachline.toFloat();

                //colour gauges based on the thickness
                const GLubyte curv_col = qMin(fabsf(ground_truths[i][j].local_thickness) * 20.0f * 255.0f, 255.0f);
                ground_truths_gauges[i][j].SetColour(255, 255 - curv_col, 255 - curv_col);

            }

            file.close();

        }

        //load abstraction distance info into gauge
        if (experimental_mode && abstraction > 0) {

            QString filename;

            switch (abstraction) {
            case 1:
                filename= QString("./input_models/") + meshnames[i] + QString("/gauge_absdist_mps.txt");
                break;
            case 2:
                filename= QString("./input_models/") + meshnames[i] + QString("/gauge_absdist_xyz.txt");
                break;
            case 3:
                filename= QString("./input_models/") + meshnames[i] + QString("/gauge_absdist_radial.txt");
                break;
            case 4:
                filename= QString("./input_models/") + meshnames[i] + QString("/gauge_absdist_crdbrd.txt");
                break;
            }

            QFile file(filename);

            if (!file.open(QIODevice::ReadOnly | QIODevice::Text)) {
                qDebug() << "File " << filename << " can't be read";
                continue;
            }

            qDebug() << "Reading" << filename << "...";

            QTextStream ifs(&file);

            for (int j=0; j<ground_truths[i].size(); ++j) {

                QString eachline = ifs.readLine();

                ground_truths[i][j].abstraction_distance = eachline.toFloat();

                //colour gauges based on the thickness
                const GLubyte curv_col = qMin(fabsf(ground_truths[i][j].abstraction_distance) * 20.0f * 255.0f, 255.0f);
                ground_truths_gauges[i][j].SetColour(255, 255 - curv_col, 255 - curv_col);

            }

            file.close();

        }

        //load abstraction angle difference info for gauge
        if (experimental_mode && abstraction > 0) {

            QString filename;

            switch (abstraction) {
            case 1:
                filename= QString("./input_models/") + meshnames[i] + QString("/gauge_absdiff_mps.txt");
                break;
            case 2:
                filename= QString("./input_models/") + meshnames[i] + QString("/gauge_absdiff_xyz.txt");
                break;
            case 3:
                filename= QString("./input_models/") + meshnames[i] + QString("/gauge_absdiff_radial.txt");
                break;
            case 4:
                filename= QString("./input_models/") + meshnames[i] + QString("/gauge_absdiff_crdbrd.txt");
                break;
            }

            QFile file(filename);

            if (!file.open(QIODevice::ReadOnly | QIODevice::Text)) {
                qDebug() << "File " << filename << " can't be read";
                continue;
            }

            qDebug() << "Reading" << filename << "...";

            QTextStream ifs(&file);

            for (int j=0; j<ground_truths[i].size(); ++j) {

                QString eachline = ifs.readLine();

                ground_truths[i][j].abstraction_angle_difference = eachline.toFloat();

                //colour gauges based on the thickness
                const GLubyte curv_col = qMin(fabsf(ground_truths[i][j].abstraction_angle_difference) * 2.5f, 255.0f);
                ground_truths_gauges[i][j].SetColour(255, 255 - curv_col, 255 - curv_col);

            }

            file.close();

        }

        //load "centroid distance" info for gauge
        if (experimental_mode) {

            QString filename = QString("./input_models/") + meshnames[i] + QString("/gauge_centdist.txt");
            QFile file(filename);

            if (!file.open(QIODevice::ReadOnly | QIODevice::Text)) {
                qDebug() << "File " << filename << " can't be read";
                continue;
            }

            qDebug() << "Reading" << filename << "...";

            QTextStream ifs(&file);

            for (int j=0; j<ground_truths[i].size(); ++j) {

                QString eachline = ifs.readLine();

                ground_truths[i][j].centroid_distance = eachline.toFloat();

                //colour gauges based on the thickness
                const GLubyte curv_col = qMin(fabsf(ground_truths[i][j].centroid_distance) * 255.0f, 255.0f);
                ground_truths_gauges[i][j].SetColour(255, 255 - curv_col, 255 - curv_col);

            }

            file.close();

        }

        //load "centroid distance" info for gauge
        if (experimental_mode) {

            QString filename = QString("./input_models/") + meshnames[i] + QString("/gauge_viewnorm.txt");
            QFile file(filename);

            if (!file.open(QIODevice::ReadOnly | QIODevice::Text)) {
                qDebug() << "File " << filename << " can't be read";
                continue;
            }

            qDebug() << "Reading" << filename << "...";

            QTextStream ifs(&file);

            for (int j=0; j<ground_truths[i].size(); ++j) {

                QString eachline = ifs.readLine();

                ground_truths[i][j].viewnorm_angle = eachline.toFloat();

                //colour gauges based on the thickness
                const GLubyte curv_col = qMin(fabsf(ground_truths[i][j].viewnorm_angle) / 90.0f * 255.0f, 255.0f);
                ground_truths_gauges[i][j].SetColour(255, 255 - curv_col, 255 - curv_col);

            }

            file.close();

        }


    }

}

void GLWidget::LoadSplashImage()
{   

    QString filename;

    if (study_task == ROTATE) {
        if (abstraction == 0) {
            filename = QString("./splash_rot.png");
        }
        else {
            filename = QString("./splash_rot_mps.png");
        }
    }
    else if (study_task == ROTATE_VIEW) {
        if (abstraction == 0) {
            filename = QString("./splash_rot_view.png");
        }
        else {
            filename = QString("./splash_rot_view_mps.png");
        }
    }

    QImage splash_image;
    if (!splash_image.load(filename)) {
        qDebug() << "GLWidget::LoadSplashImage() - warning, could not load" << filename;
    }

    splash_image_gl = convertToGLFormat(splash_image);

}

void GLWidget::DrawSplashImage()
{

    glDrawPixels(splash_image_gl.width(), splash_image_gl.height(), GL_RGBA, GL_UNSIGNED_BYTE, splash_image_gl.bits());

}

void GLWidget::AddCustomGauge(const int mesh_index)
{

    StudyData d;
    if (GetGaugeForPoint(mouse_pos, mesh_index, d)) {

        ground_truths[mesh_index].push_back(d);

        GaugeManipulator g;

        g.SetCentre(d.pos);
        g.SetNormal(d.true_normal);
        g.SetScale(gauge_scale);
        g.SetColour(0, 0, 255);

        ground_truths_gauges[mesh_index].push_back(g);

    }

    //special case for bumpy cylinder -we add the exact same gauge to the plain cylinder
    if (mesh_index == 5) {
        qDebug() << "Attention!!!  On" << meshnames[5] << "so adding a gauge to" << meshnames[4] << "as well!";
        AddCustomGauge(4);
    }

}
