#include "histogram.h"

Histogram::Histogram()
{

    lo_val = 0.0f;
    hi_val = 90.0f;
    bucket_size = 5.0f;

}

void Histogram::AddSample(const float f)
{

    labels.push_back(QString::number(samples.size()));
    samples.push_back(f);    

}

void Histogram::AddSample(const float f, const QString l)
{

    labels.push_back(l);
    samples.push_back(f);

}

void Histogram::Clear()
{

    samples.clear();
    labels.clear();

}

int Histogram::GetNumSamples() const
{
    return samples.size();
}

void Histogram::SetRange(const float lo, const float hi)
{
    lo_val = lo;
    hi_val = hi;
}

void Histogram::SetBucketSize(const float b)
{
    bucket_size = b;
}

void Histogram::OutputSamples() const
{
    for (int i=0; i<samples.size(); ++i) {
        qDebug() << samples[i];
    }
}

void Histogram::OutputSamplesToFile(const QString & filename) const
{

    QFile file(filename);
    if (!file.open(QIODevice::WriteOnly | QIODevice::Text)) {
        qDebug() << "cant write" << filename;
        return;
    }
    qDebug() << "Writing" << filename << "...";

    QTextStream ofs(&file);

    for (int i=0; i<samples.size(); ++i) {
        ofs << samples[i] << "\n";
    }

    file.close();

}

void Histogram::OutputData() const
{

    //output angle between guesses histogram data
    for (float i=lo_val; i<hi_val; i+=bucket_size) {

        int each_freq = 0;
        for (int j=0; j<samples.size(); ++j) {

            if (samples[j] >= i && samples[j] < i + bucket_size) {
                ++each_freq;
            }

        }

        qDebug() << each_freq;

    }

}

void Histogram::OutputLessThanData(const QList <float> & avg_y_list, const float max_y) const
{

    //output angle between guesses histogram data
    for (float i=lo_val; i<hi_val; i+=bucket_size) {

        int each_freq = 0;
        for (int j=0; j<samples.size(); ++j) {

            if (samples[j] >= i && samples[j] < i + bucket_size && avg_y_list[j] < max_y) {
                ++each_freq;
            }

        }

        qDebug() << i << each_freq;

    }

}

void Histogram::OutputLessThanNormalizedData(const QList <float> & avg_y_list, const float max_y) const
{

    //output angle between guesses histogram data
    for (float i=lo_val; i<hi_val; i+=bucket_size) {

        int numer_freq = 0;
        int denom_freq = 0;
        for (int j=0; j<samples.size(); ++j) {

            if (samples[j] >= i && samples[j] < i + bucket_size) {

                ++denom_freq;

                if (avg_y_list[j] < max_y) {
                    ++numer_freq;
                }

            }

        }

        if (denom_freq > 0) {
            qDebug() << i << float(numer_freq) / float(denom_freq);
        }

    }

}

void Histogram::OutputAverageData(const QList <float> & avg_y_list) const
{

    //output angle between guesses histogram data
    for (float i=lo_val; i<hi_val; i+=bucket_size) {

        int num_freq = 0;
        float avg_val = 0.0f;

        for (int j=0; j<samples.size(); ++j) {

            if (samples[j] >= i && samples[j] < i + bucket_size) {

                ++num_freq;
                avg_val += avg_y_list[j];

            }

        }

        if (num_freq > 0) {
            qDebug() << i << avg_val / float(num_freq);
        }
        else {
            qDebug() << i << "0";
        }

    }

}

float Histogram::GetMean() const
{

    const int s = samples.size();
    float mean = 0.0f;

    for (int i=0; i<samples.size(); ++i) {

        mean += samples[i];

    }

    mean /= float(s);

    return mean;

}

float Histogram::GetStdDev() const
{

    const float mean = GetMean();
    const int s = samples.size();
    float sd = 0.0f;

    for (int i=0; i<s; ++i) {

        float diff = samples[i] - mean;
        sd += diff * diff;

    }

    sd /= float(s);
    sd = sqrtf(sd);

    return sd;

}

float Histogram::GetMin() const
{

    float min_val = FLT_MAX;

    for (int i=0; i<samples.size(); ++i) {
        min_val = qMin(min_val, samples[i]);
    }

    return min_val;

}

float Histogram::GetMax() const
{

    float max_val = -FLT_MAX;

    for (int i=0; i<samples.size(); ++i) {
        max_val = qMax(max_val, samples[i]);
    }

    return max_val;

}

QList <QString> Histogram::GetOutliers(const int n_std_dev) const
{

    const float mean = GetMean();
    const float std_dev = GetStdDev();

    QList <QString> outliers;

    for (int i=0; i<samples.size(); ++i) {

        if (samples[i] > mean + (std_dev * n_std_dev) ||
                samples[i] < mean - (std_dev * n_std_dev)) {

            outliers.push_back(labels[i]);

        }

    }

    return outliers;

}
