#ifndef HISTOGRAM_H
#define HISTOGRAM_H

#include <QtCore>
#include <cfloat>

class Histogram
{

public:

    Histogram();

    void AddSample(const float f);
    void AddSample(const float f, const QString l);
    void Clear();

    int GetNumSamples() const;

    void SetRange(const float lo, const float hi);
    void SetBucketSize(const float b);

    void OutputSamples() const;
    void OutputSamplesToFile(const QString & filename) const;
    void OutputData() const;
    void OutputLessThanData(const QList <float> & avg_y_list, const float max_y) const;
    void OutputLessThanNormalizedData(const QList <float> & avg_y_list, const float max_y) const;
    void OutputAverageData(const QList <float> & avg_y_list) const;

    float GetMean() const;
    float GetStdDev() const;
    float GetMin() const;
    float GetMax() const;
    QList <QString> GetOutliers(const int n_std_dev) const;


private:

    QList <float> samples;
    QList <QString> labels;

    float lo_val;
    float hi_val;
    float bucket_size;

};

#endif // HISTOGRAM_H
