#include "mainwindow.h"

MainWindow::MainWindow(QWidget *parent)
    : QMainWindow(parent)
{           

    //check that the max # runs not yet been reached
    const int max_trial_runs = 10;

    TrialRecord trial_records;

    if (trial_records.Load("./records.txt")) {

        if (trial_records.GetStudyRuns() >= max_trial_runs) {

            qDebug() << "Max # runs reached, aborting";

            QMessageBox m;
            m.setText("You have reached the maximum number of " \
                      "successful runs of the study (no more than " + QString::number(max_trial_runs) +
                      " HITs per Worker ID will be accepted).");
            m.setWindowTitle("Program Exiting");
            m.exec();

            exit(0);

        }

    }



    //const int input_val = dialog.getInt(this, "PlanarCutPerception - Input HIT #", "Enter the exact participant number shown in the Mechanical Turk HIT.", 0, 0, num_turk_hits, 0, &ok);


    QString turker_id("11111111111111");
    /*
    bool ok;
    QString turker_id = QInputDialog::getText(this, "Enter Mechanical Turk Worker ID", "Type or copy/paste in your 12-14 character Worker ID from Mechanical Turk.", QLineEdit::Normal, QString(), &ok);

    turker_id = turker_id.toUpper(); //make uppercase
    turker_id = turker_id.trimmed(); //remove whitespace (may exist if copy/pasted in)

    QRegExp rx("^[A-Z0-9]{12,14}$");

    //check that an id was entered in last window, and that it matches the regular expression
    if (!ok || rx.indexIn(turker_id) < 0) {

        qDebug() << "Turker ID" << turker_id << "does not match regular expression";

        QMessageBox m;
        m.setText("You must enter your valid Mechanical Turk Worker ID to participate in the study (it consists of 12 to 14 alpha-numeric characters).");
        m.setWindowTitle("Program Exiting");
        m.exec();

        exit(0);

    }

    TurkWorkerData work_data;
    if (work_data.WorkerInFirstRun(turker_id)) {

        qDebug() << "Turker ID participated in first run";

        QMessageBox m;
        m.setText("Turker IDs which participated in the first run of the study are not allowed in this second run.");
        m.setWindowTitle("Program Exiting");
        m.exec();

        exit(0);

    }
    */

    glwidget.SetupStudy(turker_id);

    nexttrial.setText("Next Trial (Press Spacebar)");

    studydatatextedit.setReadOnly(true);

    connect(&nexttrial, SIGNAL(clicked()), &glwidget, SLOT(SetupNewTrial()));

    //if (false) { //set to false to hide that top progress bar, etc., useful for screenshot
        layout.addWidget(new QLabel("Progress:"), 0, 0);
        layout.addWidget(&progress, 0, 1);
        layout.addWidget(&nexttrial, 0, 2);
    //}
    layout.addWidget(&glwidget, 1, 0, 1, 3);

    QWidget * winwidget = new QWidget;
    winwidget->setLayout(&layout);

    setCentralWidget(winwidget);

    connect(&glwidget, SIGNAL(StartedNewTrial()), this, SLOT(UpdateProgress()));
    connect(&glwidget, SIGNAL(EnableNextTrialButton()), this, SLOT(EnableNextTrialButton()));
    connect(&glwidget, SIGNAL(DisableNextTrialButton()), this, SLOT(DisableNextTrialButton()));

    setWindowTitle("PlanarCutPerception");
    resize(1024, 768);

    //setup finished widget
    QVBoxLayout * finishlayout = new QVBoxLayout;

    QPushButton * copypushbutton = new QPushButton("Click to copy data to clipboard");

    connect(copypushbutton, SIGNAL(clicked()), &glwidget, SLOT(CopyStudyDataToClipboard()));

    finishlayout->addWidget(new QLabel("Success: Data copied to clipboard.  Paste data into HIT for Worker " + turker_id + " to receive payment.  Thank you!"));
    finishlayout->addWidget(copypushbutton);
    finishlayout->addWidget(&studydatatextedit);
    studydatatextedit.setEnabled(false);

    finishedwidget.setLayout(finishlayout);

    failedwidget.setText("Failed: Gauges do not meet minimum for precision - the result will not be accepted on Mechanical Turk.  Re-run the study and take care to be more precise.");

    UpdateProgress();

}

MainWindow::~MainWindow()
{
    
}

void MainWindow::keyPressEvent(QKeyEvent *event)
{

    glwidget.keyPressEvent(event);

}

void MainWindow::keyReleaseEvent(QKeyEvent *event)
{

    glwidget.keyReleaseEvent(event);

}

void MainWindow::UpdateProgress()
{

    progress.setRange(0, glwidget.GetNumTrials());
    progress.setValue(glwidget.GetCurTrial());

    //if progress finished, show the final message saying copy/paste this into turk

    if (glwidget.StudyFinished()) {

        if (glwidget.StudySuccessful()) {

            layout.addWidget(&finishedwidget, 2, 0, 1, 3);

            studydatatextedit.setText(glwidget.GetStudyData());

        }
        else {

            layout.addWidget(&failedwidget, 2, 0, 1, 3);

        }

    }


}

void MainWindow::EnableNextTrialButton()
{

    nexttrial.setEnabled(true);

}

void MainWindow::DisableNextTrialButton()
{

    nexttrial.setEnabled(false);

}
