#include "normmanipulator.h"

NormManipulator::NormManipulator() :
    cent(0, 0, 0),
    scale(0.5f),
    scalecontrolfac(0.8f),
    state(PM_NONE),
    mouseover(PM_NONE)
{

    rot1col[0] = 1;
    rot1col[1] = 1;
    rot1col[2] = 1;

    rot2col[0] = 2;
    rot2col[1] = 2;
    rot2col[2] = 2;

    rotncol[0] = 3;
    rotncol[1] = 3;
    rotncol[2] = 3;

    scalecol[0] = 4;
    scalecol[1] = 4;
    scalecol[2] = 4;

    rot1drawcol[0] = 50;
    rot1drawcol[1] = 255;
    rot1drawcol[2] = 50;

    rot2drawcol[0] = 50;
    rot2drawcol[1] = 255;
    rot2drawcol[2] = 50;

    rotndrawcol[0] = 50;
    rotndrawcol[1] = 255;
    rotndrawcol[2] = 50;

    scaledrawcol[0] = 50;
    scaledrawcol[1] = 255;
    scaledrawcol[2] = 50;

    alwaysdrawcol[0] = 220;
    alwaysdrawcol[1] = 220;
    alwaysdrawcol[2] = 220;

    time.start();

}

void NormManipulator::mousePressEvent(int x, int y)
{

    state = mouseover;

    last_x = x;
    last_y = y;

}

void NormManipulator::mouseMoveEvent(int x, int y, OrbitCamera & cam)
{

    UpdatePickGL(x, y);

    vec3 udDisplace = cam.Up();
    vec3 lrDisplace = udDisplace ^ (cam.ViewDir());

    udDisplace.normalize();
    lrDisplace.normalize();

    int winWidth = GLutils::GetWindowWidth();
    int winHeight = GLutils::GetWindowHeight();

    float rotDelta = (x - last_x) + (y - last_y);

    switch (state) {

    case PM_ROT1:
    {

        p.RotateFrame(p.B(), rotDelta);

    }

        break;

    case PM_ROT2:
    {

        p.RotateFrame(p.T(), rotDelta);

    }

        break;

    case PM_ROTN:
    {

        p.RotateFrame(p.N(), rotDelta);

    }

        break;

    case PM_SCALE:
    {

    }

        break;

    default:

        break;

    }

    last_x = x;
    last_y = y;

}

void NormManipulator::UpdatePickGL(int x, int y)
{

    //glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    DrawPickGL();

    unsigned char r, g, b;
    GLutils::ReadPixelColor_BackBuffer(x, y, r, g, b);

    if (r == rot1col[0] && g == rot1col[1] && b == rot1col[2]) {
        mouseover = PM_ROT1;
    }
    else if (r == rot2col[0] && g == rot2col[1] && b == rot2col[2]) {
        mouseover = PM_ROT2;
    } 
    else if (r == rotncol[0] && g == rotncol[1] && b == rotncol[2]) {
        mouseover = PM_ROTN;
    }
    else if (r == scalecol[0] && g == scalecol[1] && b == scalecol[2]) {
        mouseover = PM_SCALE;
    }
    else {
        mouseover = PM_NONE;
    }

}

void NormManipulator::mouseReleaseEvent(int x, int y)
{

    state = PM_NONE;

    last_x = x;
    last_y = y;

}

void NormManipulator::DrawGL()
{

    UpdateVerts();

    float whiteval = ((sinf(time.elapsed() / 50.0f)) + 1.0f) * 0.25f;
    vec3 whitecol = vec3(.5f + whiteval, .5f + whiteval, .75f + whiteval);   

    float thick_inner = 3.0f;
    float thick_outer = 5.0f;   

    //draw shading plane
    /*
    GLutils::EnableBlending();

    glColor4f(1, 1, 1, 0.5f);

    glBegin(GL_QUADS);
    glVertex3f(v1[0], v1[1], v1[2]);
    glVertex3f(v2[0], v2[1], v2[2]);
    glVertex3f(v3[0], v3[1], v3[2]);
    glVertex3f(v4[0], v4[1], v4[2]);
    glEnd();

    GLutils::DisableBlending();
    */

    //glDepthMask(GL_FALSE);

    //rotN control
    glLineWidth(thick_inner);
    if (state == PM_ROTN || (state == PM_NONE && mouseover == PM_ROTN)) {
        glColor3f(whitecol[0], whitecol[1], whitecol[2]);
    }
    else {
        glColor3ub(rotndrawcol[0], rotndrawcol[1], rotndrawcol[2]);
    }
    DrawRotNLines();

    glLineWidth(thick_outer);
    glColor3ub(0, 0, 0);
    DrawRotNLines();

    //rot1 control
    glLineWidth(thick_inner);
    if (state == PM_ROT1 || (state == PM_NONE && mouseover == PM_ROT1)) {
        glColor3f(whitecol[0], whitecol[1], whitecol[2]);
    }
    else {
        glColor3ub(rot1drawcol[0], rot1drawcol[1], rot1drawcol[2]);
    }
    DrawRot1Lines();

    glLineWidth(thick_outer);
    glColor3ub(0, 0, 0);
    DrawRot1Lines();   

    //rot2 control
    glLineWidth(thick_inner);
    if (state == PM_ROT2 || (state == PM_NONE && mouseover == PM_ROT2)) {
        glColor3f(whitecol[0], whitecol[1], whitecol[2]);
    }
    else {
        glColor3ub(rot2drawcol[0], rot2drawcol[1], rot2drawcol[2]);
    }
    DrawRot2Lines();

    glLineWidth(thick_outer);
    glColor3ub(0, 0, 0);
    DrawRot2Lines();   

    //draw scale control
    /*
    glLineWidth(thick_inner);
    if (state == PM_SCALE || (state == PM_NONE && mouseover == PM_SCALE)) {
        glColor3f(whitecol[0], whitecol[1], whitecol[2]);
    }
    else {
        glColor3ub(scaledrawcol[0], scaledrawcol[1], scaledrawcol[2]);
    }
    DrawScaleLines();

    glLineWidth(thick_outer);
    glColor3ub(0, 0, 0);
    DrawScaleLines();   
    */

    //always lines
    glLineWidth(thick_inner);
    glColor3f(alwaysdrawcol[0], alwaysdrawcol[1], alwaysdrawcol[2]);
    DrawAlwaysLines();

    glLineWidth(thick_outer);
    glColor3ub(0, 0, 0);
    DrawAlwaysLines();

    //glDepthMask(GL_TRUE);

    /*
    //visual debugging
    glColor3f(1, 0, 0);
    glBegin(GL_POINTS);
    glVertex3f(transint[0], transint[1], transint[2]);
    glEnd();
    transplane.DrawGL();
    */

    glLineWidth(1.0f);
    glColor3f(1,1,1);

}

void NormManipulator::DrawPickGL()
{

    UpdateVerts();

    //rotn lines
    glLineWidth(10.0f);
    glColor3ub(rotncol[0], rotncol[1], rotncol[2]);

    DrawRotNLines();

    //rot1 line
    glLineWidth(10.0f);
    glColor3ub(rot1col[0], rot1col[1], rot1col[2]);

    DrawRot1Lines();

    //rot2 line
    glLineWidth(10.0f);
    glColor3ub(rot2col[0], rot2col[1], rot2col[2]);

    DrawRot2Lines();

    //scale lines
    /*
    glLineWidth(10.0f);
    glColor3ub(scalecol[0], scalecol[1], scalecol[2]);

    DrawScaleLines();   
    */

    glLineWidth(1.0f);
    glColor3f(1,1,1);

}

NormManipState NormManipulator::GetState()
{

    return state;

}

vec3 NormManipulator::GetCentre()
{
    return cent;
}

void NormManipulator::SetCentre(const vec3 & p)
{

    cent = p;    

}

void NormManipulator::SetNormal(const vec3 & n)
{

    p.SetN(n);

}

vec3 NormManipulator::GetNormal()
{

    return p.N();

}

void NormManipulator::SetScale(const float f)
{
    scale = f;
}

float NormManipulator::GetScale()
{
    return scale;
}

void NormManipulator::DrawRotNLines()
{

    glBegin(GL_LINE_LOOP);

    for (float i=0.0f; i<ALG_M2_PI; i+=0.1f) {

        const vec3 v = cent + p.T() * sinf(i) * scale * 0.1f + p.B() * cosf(i) * scale * 0.1f;
        glVertex3f(v[0], v[1], v[2]);

    }

    glEnd();

}

void NormManipulator::DrawRot1Lines()
{

    glBegin(GL_TRIANGLES);
    glVertex3f(v2[0], v2[1], v2[2]);
    glVertex3f(v4[0], v4[1], v4[2]);
    glVertex3f(v5[0], v5[1], v5[2]);

    glVertex3f(v2[0], v2[1], v2[2]);
    glVertex3f(v6[0], v6[1], v6[2]);
    glVertex3f(v3[0], v3[1], v3[2]);
    glEnd();

    if (state == PM_ROT1) {

        glBegin(GL_LINES);
        glVertex3f(v9[0], v9[1], v9[2]);
        glVertex3f(v10[0], v10[1], v10[2]);
        glEnd();

    }

}

void NormManipulator::DrawRot2Lines()
{

    glBegin(GL_TRIANGLES);
    glVertex3f(v2[0], v2[1], v2[2]);
    glVertex3f(v3[0], v3[1], v3[2]);
    glVertex3f(v4[0], v4[1], v4[2]);

    glVertex3f(v2[0], v2[1], v2[2]);
    glVertex3f(v5[0], v5[1], v5[2]);
    glVertex3f(v6[0], v6[1], v6[2]);
    glEnd();

    //rotation guides
    if (state == PM_ROT2) {

        glBegin(GL_LINES);
        glVertex3f(v7[0], v7[1], v7[2]);
        glVertex3f(v8[0], v8[1], v8[2]);
        glEnd();

    }

}

void NormManipulator::DrawTransLines()
{

    glBegin(GL_LINES);
    glVertex3f(v5[0], v5[1], v5[2]);
    glVertex3f(v6[0], v6[1], v6[2]);

    glVertex3f(v5[0], v5[1], v5[2]);
    glVertex3f(v7[0], v7[1], v7[2]);
    glEnd();

}

void NormManipulator::DrawScaleLines()
{

    /*
    glBegin(GL_LINES);
    glVertex3f(v1[0], v1[1], v1[2]);
    glVertex3f(v12[0], v12[1], v12[2]);

    glVertex3f(v1[0], v1[1], v1[2]);
    glVertex3f(v14[0], v14[1], v14[2]);

    glVertex3f(v2[0], v2[1], v2[2]);
    glVertex3f(v21[0], v21[1], v21[2]);

    glVertex3f(v2[0], v2[1], v2[2]);
    glVertex3f(v23[0], v23[1], v23[2]);

    glVertex3f(v3[0], v3[1], v3[2]);
    glVertex3f(v32[0], v32[1], v32[2]);

    glVertex3f(v3[0], v3[1], v3[2]);
    glVertex3f(v34[0], v34[1], v34[2]);

    glVertex3f(v4[0], v4[1], v4[2]);
    glVertex3f(v43[0], v43[1], v43[2]);

    glVertex3f(v4[0], v4[1], v4[2]);
    glVertex3f(v41[0], v41[1], v41[2]);
    glEnd();
    */

}

void NormManipulator::DrawAlwaysLines()
{

    glBegin(GL_LINES);
    //stem
    glVertex3f(v1[0], v1[1], v1[2]);
    glVertex3f(v2[0], v2[1], v2[2]);

    //arrow head base
    glVertex3f(v2[0], v2[1], v2[2]);
    glVertex3f(v3[0], v3[1], v3[2]);

    glVertex3f(v2[0], v2[1], v2[2]);
    glVertex3f(v4[0], v4[1], v4[2]);

    glVertex3f(v2[0], v2[1], v2[2]);
    glVertex3f(v5[0], v5[1], v5[2]);

    glVertex3f(v2[0], v2[1], v2[2]);
    glVertex3f(v6[0], v6[1], v6[2]);


    glVertex3f(v3[0], v3[1], v3[2]);
    glVertex3f(v4[0], v4[1], v4[2]);

    glVertex3f(v4[0], v4[1], v4[2]);
    glVertex3f(v5[0], v5[1], v5[2]);

    glVertex3f(v5[0], v5[1], v5[2]);
    glVertex3f(v6[0], v6[1], v6[2]);

    glVertex3f(v6[0], v6[1], v6[2]);
    glVertex3f(v3[0], v3[1], v3[2]);

    glEnd();

}

void NormManipulator::UpdateVerts()
{

    vec3 tscale = p.T() * scale;
    vec3 bscale = p.B() * scale;
    float scalefac = scalecontrolfac;

    v1 = cent;
    v2 = cent + p.N() * scale;
    v3 = cent + (p.N() * 0.8f - p.T() * 0.1f - p.B() * 0.1f) * scale;
    v4 = cent + (p.N() * 0.8f + p.T() * 0.1f - p.B() * 0.1f) * scale;
    v5 = cent + (p.N() * 0.8f + p.T() * 0.1f + p.B() * 0.1f) * scale;
    v6 = cent + (p.N() * 0.8f - p.T() * 0.1f + p.B() * 0.1f) * scale;
    v7 = cent - p.T() * scale * 0.2f;
    v8 = cent + p.T() * scale * 0.2f;
    v9 = cent - p.B() * scale * 0.2f;
    v10 = cent + p.B() * scale * 0.2f;

}
