#ifndef NORMMANIPULATOR_H
#define NORMMANIPULATOR_H

#include <QtOpenGL>

#include "algebra3.h"
#include "glutils.h"
#include "orbitcamera.h"

enum NormManipState
{
    PM_NONE,
    PM_ROTN,
    PM_ROT1,
    PM_ROT2,
    PM_SCALE
};

class NormManipulator
{

public:

    NormManipulator();

    void mousePressEvent(int x, int y);
    void mouseMoveEvent(int x, int y, OrbitCamera & cam);
    void mouseReleaseEvent(int x, int y);

    void DrawGL();

    NormManipState GetState();

    vec3 GetCentre();
    void SetCentre(const vec3 & p);

    void SetNormal(const vec3 & n);
    vec3 GetNormal();

    void SetScale(const float f);
    float GetScale();

private:

    void DrawPickGL();
    void UpdatePickGL(int x, int y);

    void DrawRotNLines();
    void DrawRot1Lines();
    void DrawRot2Lines();
    void DrawTransLines();
    void DrawScaleLines();

    void DrawAlwaysLines();

    void UpdateVerts();

    TNBFrame p; //plane implemented as tnb frame (normal is N)
    vec3 cent;
    float scale;
    float scalecontrolfac;

    GLubyte rot1col[3];
    GLubyte rot2col[3];
    GLubyte rotncol[3];
    GLubyte scalecol[3];    

    GLubyte rot1drawcol[3];
    GLubyte rot2drawcol[3];
    GLubyte rotndrawcol[3];
    GLubyte scaledrawcol[3];

    GLubyte alwaysdrawcol[3];

    int last_x;
    int last_y;

    NormManipState state;
    NormManipState mouseover;

    QTime time;

    vec3 v1;
    vec3 v2;
    vec3 v3;
    vec3 v4;
    vec3 v5;
    vec3 v6;
    vec3 v7;
    vec3 v8;
    vec3 v9;
    vec3 v10;

    //for visual debugging
    //Plane transplane;
    //vec3 transint;

};

#endif // NORMMANIPULATOR_H
