#ifndef ORBITCAMERA_H
#define ORBITCAMERA_H

#include <QtOpenGL>
#include <GL/glu.h>

#include "algebra3.h"

class OrbitCamera
{
public:

    OrbitCamera();

    void DrawGL() const;
    void DrawGL_Ortho() const;
    void DrawGL_3DOrtho() const;
    void Update();

    void Orbit(float dx, float dy);
    void Pan(float dx, float dy);
    void ZoomIn();
    void ZoomOut();

    vec3 SphericalToCartesian(float theta, float phi, float r) const;
    vec3 Up() const;
    vec3 Eye() const;
    vec3 LookAt() const;
    void SetLookAt(const vec3 & p);

    float Theta() const;
    void SetTheta(float f);
    float Phi() const;
    void SetPhi(float f);

    vec3 ViewDir() const;
    float ViewDist() const;
    void SetViewDist(float f);

    void SetWinWidth(int i);
    void SetWinHeight(int i);
    void SetCamWidth(float f);
    float CamWidth() const;

    void SetInterpCamera(OrbitCamera & cam, float duration);
    float InterpTime() const;
    template <typename T>
    T InterpValue_Linear(T v1, T v2, float interp) const;
    template <typename T>
    T InterpValue_Cosine(T v1, T v2, float interp) const;

private:

    double neardist;
    double fardist;

    int win_width;
    int win_height;
    float cam_width;

    float theta;
    float phi;
    float viewdist;
    vec3 eye;
    vec3 up;
    vec3 lookat;

    bool interp_active;
    QTime interp_time;
    float interp_duration;
    OrbitCamera * interp_cam_start;
    OrbitCamera * interp_cam_end;

};

#endif // ORBITCAMERA_H
