#ifndef PLANE_H
#define PLANE_H

#include <QtOpenGL>

#include "algebra3.h"
#include "tnbframe.h"
#include "eigen.h"

class Plane : public TNBFrame
{

public:

    Plane();

    void SetPts(const vec3 & p1, const vec3 & p2, const vec3 & p3);
    void SetD(const float val);
    float D() const;

    vec3 GetCentre() const;

    void DrawGL() const;
    void DrawLoopGL(const vec3 & centre, const float t_size, const float b_size) const;
    void DrawQuadGL(const vec3 & centre, const float t_size, const float b_size) const;
    void DrawBoundingBoxGL() const;

    float PointDistance(const vec3 & p) const;
    bool RayIntersect(const vec3 & p0, const vec3 & dir, vec3 & intersect) const;
    bool LineSegmentIntersect(const vec3 & p0, const vec3 & p1, vec3 & intersect) const;
    bool LineSegmentIntersect(const vec3 & p0, const vec3 & p1, vec3 & intersect, float & u) const;

    void BestFitPlane(const QList <vec3> & points, vec3 & eigen1, vec3 & eigen2, vec3 & eigen3, vec3 & centroid);

    vec2 GetPoint2D(const vec3 & p) const;
    vec3 GetPoint3D(const vec2 & p) const;

    vec3 MirrorPoint(const vec3 & p) const;

    void SetBoundingBox(const float min_t, const float min_b, const float max_t, const float max_b);

private:

    //distance from origin (n*x - d = 0) defines plane
    float d;

    float bbox_min_t;
    float bbox_min_b;
    float bbox_max_t;
    float bbox_max_b;

};

#endif // PLANE_H
