#include "scatterplot.h"

Scatterplot::Scatterplot() :
    model_a(0.0f),
    model_b(0.0f),
    model_recompute(false)
{
}

void Scatterplot::AddSample(const float x, const float y)
{

    labels.push_back(QString::number(sample_x.size()));
    sample_x.push_back(x);
    sample_y.push_back(y);

    model_recompute = true;

}

void Scatterplot::AddSample(const float x, const float y, const QString & l)
{

    labels.push_back(l);
    sample_x.push_back(x);
    sample_y.push_back(y);

    model_recompute = true;

}

void Scatterplot::Clear()
{

    labels.clear();
    sample_x.clear();
    sample_y.clear();

    model_recompute = true;

}

float Scatterplot::GetMeanX() const
{

    const int s = sample_x.size();
    float mean = 0.0f;

    for (int i=0; i<sample_x.size(); ++i) {

        mean += sample_x[i];

    }

    mean /= float(s);

    return mean;

}

float Scatterplot::GetMeanY() const
{

    const int s = sample_y.size();
    float mean = 0.0f;

    for (int i=0; i<sample_y.size(); ++i) {

        mean += sample_y[i];

    }

    mean /= float(s);

    return mean;

}

float Scatterplot::GetVarianceX() const
{

    const float mean_x = GetMeanX();

    float var_x = 0.0f;

    for (int i=0; i<sample_x.size(); ++i) {

        var_x += (sample_x[i] - mean_x) * (sample_x[i] - mean_x);

    }

    return var_x;

}

float Scatterplot::GetVarianceY() const
{

    const float mean_y = GetMeanY();

    float var_y = 0.0f;

    for (int i=0; i<sample_y.size(); ++i) {

        var_y += (sample_y[i] - mean_y) * (sample_y[i] - mean_y);

    }

    return var_y;

}

float Scatterplot::GetCovarianceXY() const
{

    const float mean_x = GetMeanX();
    const float mean_y = GetMeanY();

    float covar_xy = 0.0f;

    for (int i=0; i<sample_x.size(); ++i) {

        covar_xy += (sample_x[i] - mean_x) * (sample_y[i] - mean_y);

    }

    return covar_xy;

}

float Scatterplot::GetResidualSumOfSquares()
{

    ComputeLinearModel();

    float res_sum_sq = 0.0f;

    for (int i=0; i<sample_x.size(); ++i) {

        const float f_x = sample_x[i] * model_a + model_b;

        res_sum_sq += (sample_y[i] - f_x) * (sample_y[i] - f_x);

    }

    return res_sum_sq;

}

float Scatterplot::GetRSquared()
{
    return 1.0f - GetResidualSumOfSquares() / GetVarianceY();
}

void Scatterplot::OutputData() const
{

    for (int i=0; i<sample_x.size(); ++i) {
        qDebug() << sample_x[i] << sample_y[i];
    }

}

void Scatterplot::OutputLinearRegression()
{

    ComputeLinearModel();

    const int deg_free = sample_x.size() - 2;
    const float r_squared = GetRSquared();
    const float test_statistic = deg_free * r_squared / (1 - r_squared);

    qDebug() << "a" << model_a << "b" << model_b << "R-squared:" << r_squared << "test_statistic:" << test_statistic;

}

void Scatterplot::ComputeLinearModel()
{

    if (!model_recompute) {
        return;
    }

    //solving regression via least squares for equation y = a*x + b
    model_a = GetCovarianceXY() / GetVarianceX(); //slope
    model_b = GetMeanY() - model_a * GetMeanX(); //y-intercept
    model_recompute = false;
}

void Scatterplot::GetLinearModelParameters(float & a, float & b) const
{

    a = model_a;
    b = model_b;

}
