#include "tnbframe.h"

TNBFrame::TNBFrame() :
    t(1, 0, 0),
    n(0, 1, 0),
    b(0, 0, 1)
{

}

vec3 TNBFrame::N() const
{
    return n;
}

vec3 TNBFrame::T() const
{
    return t;
}

vec3 TNBFrame::B() const
{
    return b;
}

vec3 TNBFrame::GetOrthoVec(const vec3 & v)
{  

    //preference to Y-up
    vec3 cross_vec;

    if (fabs(v * vec3(0, 1, 0)) > 0.9f) {

        cross_vec = vec3(0, 0, 1) ^ v;
        cross_vec.normalize();
        cross_vec = cross_vec ^ v;

    }
    else {

        cross_vec = vec3(0, 1, 0) ^ v;
        cross_vec.normalize();
        cross_vec = v ^ cross_vec;

    }

    cross_vec.normalize();

    return cross_vec;
}

void TNBFrame::SetT(const vec3 & v)
{

    t = v;
    t.normalize();

    n = GetOrthoVec(t);

    b = t ^ n;
    b.normalize();

}

void TNBFrame::SetN(const vec3 & v)
{

    n = v;
    n.normalize();

    b = GetOrthoVec(n);

    t = n ^ b;
    t.normalize();

}



void TNBFrame::SetB(const vec3 & v)
{

    b = v;
    b.normalize();

    t = GetOrthoVec(b);

    n = b ^ t;
    n.normalize();

}

void TNBFrame::SetTNB(const vec3 & vt, const vec3 & vn, const vec3 & vb)
{

    t = vt;
    n = vn;
    b = vb;

    t.normalize();
    n.normalize();
    b.normalize();

}

void TNBFrame::RotateFrame(const vec3 & axis, const float angledeg)
{

    mat3 rotMat = rotation3D(axis, angledeg);

    t = rotMat * t;
    n = rotMat * n;
    b = rotMat * b;

}

