#include "trialrecord.h"

TrialRecord::TrialRecord()
{

    study_runs = 0;
    qsrand(QDateTime::currentDateTime().toTime_t()); //seed is seconds since jan 1, 1970
    //qDebug() << QDateTime::currentDateTime().toTime_t();

}

void TrialRecord::Create(const int num_lists, const int num_records)
{

    records.clear();

    records.resize(num_lists);

    for (int i=0; i<num_lists; ++i) {
        records[i] = QVector <int> (num_records, 0);
    }

}

bool TrialRecord::Load(const QString & filename)
{

    records.clear();

    QFile file(filename);

    if (!file.open(QIODevice::ReadOnly | QIODevice::Text)) {
        qDebug() << "TrialRecord::Load() - File " << filename << " can't be opened for reading";
        return false;
    }

    QTextStream ifs(&file);

    QStringList eachline = ifs.readLine().split(" ");
    if (eachline.size() != 1) {
        qDebug() << "TrialRecord::Load() - File " << filename << " has invalid format (expected study_runs)";
        file.close();
        return false;
    }

    study_runs = eachline[0].toInt();

    eachline = ifs.readLine().split(" ");
    if (eachline.size() != 2) {
        qDebug() << "TrialRecord::Load() - File " << filename << " has invalid format (expected num models/gauges)";
        file.close();
        return false;
    }

    const int num_lists = eachline[0].toInt();
    const int num_records = eachline[1].toInt();

    Create(num_lists, num_records);

    while (!ifs.atEnd()) {

        eachline = ifs.readLine().split(" ");

        if (eachline.size() != 3) {
            qDebug() << "TrialRecord::Load() - File " << filename << " has invalid format (expected gauge count)";
            qDebug() << "  Bad line (tokenized): " << eachline;
            file.close();
            return false;
        }

        const int list_ind = eachline[0].toInt();
        const int record_ind = eachline[1].toInt();
        const int val = eachline[2].toInt();

        records[list_ind][record_ind] = val;

    }

    file.close();

    return true;

}

bool TrialRecord::Save(const QString & filename) const
{

    QFile file(filename);

    if (!file.open(QIODevice::WriteOnly | QIODevice::Text)) {
        qDebug() << "TrialRecord::Save() - File " << filename << " can't be opened for writing";
        return false;
    }

    QTextStream ofs(&file);

    if (records.empty()) {

        ofs << "0\n";
        ofs << "0 0\n";

    }
    else {

        ofs << study_runs << "\n";
        ofs << records.size() << " " << records.first().size() << "\n";

        for (int i=0; i<records.size(); ++i) {
            for (int j=0; j<records[i].size(); ++j) {
                if (records[i][j] > 0) {
                    ofs << i << " " << j << " " << records[i][j] << "\n";
                }
            }
        }

    }

    file.close();

    return true;

}

int TrialRecord::ChooseMinimalRandom(const int list_index)
{

    int min_val = INT_MAX;

    //get min val in this set
    for (int i=0; i<records[list_index].size(); ++i) {
        min_val = qMin(records[list_index][i], min_val);
    }

    QList <int> min_set;

    //get all indexes to min values
    for (int i=0; i<records[list_index].size(); ++i) {
        if (records[list_index][i] == min_val) {
            min_set.push_back(i);
        }
    }

    //choose an index at random
    const int rand_val = qrand() % min_set.size();
    const int rand_min_index = min_set[rand_val];

    //increment at that random minimal index
    ++records[list_index][rand_min_index];

    return rand_min_index;

}

int TrialRecord::GetStudyRuns()
{

    return study_runs;

}

void TrialRecord::SetStudyRuns(const int s)
{

    study_runs = s;

}
