#ifndef TURKWORKERDATA_H
#define TURKWORKERDATA_H

#include <QtCore>
#include <QtOpenGL>

#include "studydata.h"
#include "histogram.h"
#include "scatterplot.h"
#include "algebra3.h"

struct TurkWorker {

    QString id; //id has a set of hit_times
    int condition;
    QList <QString> times; //hit_time has a set of gauges
    QList <QList <StudyData> > datas;

};

class TurkWorkerData
{

public:

    TurkWorkerData();

    void AddWorker(const QString & id, const int condition);
    bool HasWorker(const QString & id);
    bool IsWorkerOutlier(const QString & id);
    int GetNumWorkers();
    QList <QString> GetWorkersWithCondition(const int condition);

    void AddWorkerHit(const QString & id, const QString & hit_time);
    bool HasWorkerHit(const QString & id, const QString & hit_time);
    void RemoveWorkerHit(const QString & id, const QString & hit_time);
    int GetWorkerNumHits(const QString & id);
    int GetWorkerNumCamManipulations(const QString & id);
    QList <QString> GetWorkerHits(const QString & id);
    float GetWorkerAvgError(const QString & id);
    float GetWorkerAvgPersistence(const QString & id);

    QList <StudyData> GetWorkerHitData(const QString & id, const QString & hit_time);

    void AddWorkerHitGauge(const QString & id, const QString & hit_time, const StudyData & d);

    void Clear();

    void ComputeDataPoints(const QVector <QList <StudyData> > & ground_truths, const int nModel, const int nSeq, const int nCond);

    int GetDataPointNumSamples(const int mod, const int seq, const int cond);
    float GetDataPointAvgError(const int mod, const int seq, const int cond);
    float GetDataPointStdDevError(const int mod, const int seq, const int cond);
    float GetDataPointAvgPersistence(const int mod, const int seq, const int cond);
    float GetDataPointAvgDuration(const int mod, const int seq, const int cond);


    void OutputStatistics(const QList <QString> & model_names);

    void DrawAllSamples(const int condition, const int model, const float gauge_size);

    static bool WorkerInFirstRun(const QString & id);

private:

    int GetWorkerIndex(const QString & id);
    int GetWorkerHitIndex(const int worker_index, const QString & hit_time);

    QList <TurkWorker> workers;
    int nM; //# models
    int nS; //# gauges
    int nC; //# conditions    
    QVector <QVector <QVector <int> > > gauge_samples;
    QVector <QVector <QVector <QList <float> > > > gauge_errors;
    QVector <QVector <QVector <float> > > gauge_avgerror;
    QVector <QVector <QVector <float> > > gauge_avgpersis;
    QVector <QVector <QVector <float> > > gauge_avgduration;

    static int num_first_run_workers;
    static QString first_run_workers [];

    static int num_outlier_workers;
    static QString outlier_workers [];

};

#endif // TURKWORKERDATA_H
